#!/usr/bin/python
# coding: utf-8

import os
import sys
from PyQt5 import QtGui, QtCore, QtWidgets
from typing import Dict

from chemscad.log import MyLog


class ChoiceModule(QtWidgets.QDialog):

    """
    Create a modal window to display the available modules. The user will
    choose through a combo box. Will open a window for the chosen module
    """

    def __init__(self, parent=None):

        super(ChoiceModule, self).__init__(parent)

        # modal doesnt let the user to interact with main window while this shows
        self.setModal(True)

        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

        self.parent = parent

        if parent is None:
            self.l = MyLog("activity.log")
            self.test = True
        else:
            self.l = self.parent.l
            self.options = self.parent.options
            self.test = False

        self.dic_modules = self.listModules()
        self.l.info(f"MODS: {self.dic_modules}")

        self.initUI()
        self.defineSlots()

    def defineSlots(self):

        """Establish the slots"""

        # Choice made, display window module
        self.ok_button.clicked.connect(self.openForm)

    def openForm(self):

        """Import the available modules"""

        # Get module's name from combo box
        module_name = self.combo_choice.currentText()
        self.l.debug(f"Module {module_name} chosen")

        module_dir = os.path.join("gui_modules", self.dic_modules[module_name])

        # Add module's directory to the import path
        sys.path.insert(0, module_dir)

        # Import module
        module = __import__(self.dic_modules[module_name])

        # Build the ChemModule class contained in the module
        # Display the module's window
        if self.test:
            chem_module = module.ChemModule(self)
        else:
            chem_module = module.ChemModule(self.parent)

        self.close()

    def listModules(self) -> Dict[str, str]:

        """
        Build the list of modules by getting names from all name.txt files
        in module_* folders
        """

        if not self.test:
            return self.parent.listModules()
        else:
            return dict()

    def initUI(self):

        """Handles the display"""

        self.setWindowTitle("Add a new module")
        #self.setMinimumSize(200, 50)

        self.combo_choice = QtWidgets.QComboBox(self)

        # Add module name to list in combo box
        self.combo_choice.addItems(self.dic_modules)
        # I hardcode it to display the reactor by default, because it is the most used
        self.combo_choice.setCurrentIndex(list(self.dic_modules).index('Reactor'))

        self.ok_button = QtWidgets.QPushButton("OK", self)

        # ----------------- ASSEMBLING ----------------------------------------

        self.vbox_global = QtWidgets.QVBoxLayout(self)
        self.vbox_global.addWidget(self.combo_choice)
        self.vbox_global.addWidget(self.ok_button)

        self.setLayout(self.vbox_global)
        self.setMinimumSize(self.sizeHint())
        self.show()


if __name__ == "__main__":

    app = QtWidgets.QApplication(sys.argv)
    obj = ChoiceModule()
    sys.exit(app.exec_())



#!/usr/bin/python
# coding: utf-8

# Define constants for OpenSCAD's rendering
# https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$fa,_$fs_and_$fn
fn = 0
fs = 0.5

#HPD - 21/11/19 - changing colours of modules in ChemCAD for consistency, more flat
colours_modules = {
    "Reactor": "#4285F4",
    "Filter reactor": "#F4B400",
    "Floating filter reactor": "#DB4437",
    "Double filter reactor": "#0F9D58",
    "S connector": "#dddddd",
    "Siphon connector": "#aaaaaa",
    "Tube connector": "#e4be9e",
    "Flow reactor": "#2ca25f",
}

colours_module_type_short = {
    "DS": colours_modules["Siphon connector"],
    "FR": colours_modules["Filter reactor"],
    "FW": colours_modules["Flow reactor"],
    "FFR": colours_modules["Floating filter reactor"],
    "DFR": colours_modules["Double filter reactor"],
    "R": colours_modules["Reactor"],
    "S": colours_modules["S connector"],
    "TC": colours_modules["Tube connector"],
}

if __name__ == "__main__":
    pass



#!/usr/bin/python
# coding: utf-8

import pytest
import numpy as np
import itertools
import subprocess as sp
import trimesh
import os
import solid

from ccad.log import MyLog
from ccad.reactor import ReactorCAD as rea
from ccad.top_in import TopInlet
from ccad.exceptions import *
import ccad.constants as cst


l = MyLog("output_tests_reactor.log", mode="w")

l.debug("---------------------- START NEW RUN OF TESTS ----------------------")
l.debug("\n")


def logAssert(test, msg):

    """
    Function to log the result of an assert
    http://stackoverflow.com/questions/24892396/py-test-logging-messages-and-test-results-assertions-into-a-single-file
    """

    if not test:
        l.error(msg)
        assert test, msg


@pytest.fixture()
def all_reactors() -> list:

    """
    Fixture:
    return all types of reactors, based on permutations of tops and bottoms
    """

    list_reactors = []

    tops = rea.TOPS
    bottoms = rea.BOTTOMS

    # Generate a list of all possible permutations
    for couple in itertools.product(tops, bottoms):
        list_reactors.append((couple[0], couple[1]))

    l.debug("List of reactors generated:\n")
    l.debug(list_reactors)

    return list_reactors


def test_error_handling():

    """Reactor class must crash if provided with invalid parameters"""

    l.debug("Testing error handling")

    # Raise exception if volume is 0
    with pytest.raises(ValueError):
        rea(0, rea.T_CLOSED_ROUND, rea.B_ROUND_IN_O)
        l.error("Reactor class: no exception raised when invalid volume")

    # Raise exception if volume is < 0
    with pytest.raises(ValueError):
        rea(-10, rea.T_CLOSED_ROUND, rea.B_ROUND_IN_O)
        l.error("Reactor class: no exception raised when invalid volume")

    # Raise exception when setting volume to -30 mL
    with pytest.raises(ValueError):
        r = rea(20, rea.T_CLOSED_ROUND, rea.B_ROUND)
        r.volume = -30
        l.error("Reactor: no exception raised when setting invalid volume")


def test_walls():

    """Tests for wall thickness (property)"""

    l.debug("Testing walls")

    r = rea(20, rea.T_CLOSED_ROUND, rea.B_ROUND_IN_O)

    # Test if walls thickness returned has valid value
    logAssert(r.walls == 3, "Wrong walls thickness returned")

    r.walls = 4

    # Test if walls thickness returned has valid value
    logAssert(r.walls == 4, "Wrong walls thickness after change")

    h = r.outputs["default"].height

    # Changing wall thickness should update dafault output height
    logAssert(h == 5.5, "Default output at wrong h {} after changing walls".format(h))

    # Walls must be at least 3 mm if threaded top
    r = rea(20, rea.T_2_QUA_INCH, rea.B_ROUND_IN_O, walls=2)
    logAssert(r.walls == 3, "Wrong walls thickness w/ threaded top")

    # Walls must be at least 3mm if ext outlet
    r = rea(20, rea.T_CLOSED_ROUND, rea.B_ROUND_EX_O)
    r.walls = 2

    logAssert(r.walls == 3, "Wrong walls thickness w/ ext outlet")


def test_volume():

    """Tests for volume (property)"""

    l.debug("Testing volume")

    r = rea(20, rea.T_CLOSED_ROUND, rea.B_ROUND)

    # Test if volume returned has proper value (should return volume in mL)
    logAssert(r.volume == 20.0, "Wrong volume returned")

    r.volume = 30

    # Test if volume returned has proper value, after modifying it
    logAssert(r.volume == 30.0, "Wrong volume returned after modif")

    logAssert(
        np.isclose(r._r, 12.79, 1e-2), "Wrong r after changing volume {}".format(r._r)
    )
    logAssert(
        np.isclose(r._h, 41.37, 1e-2), "Wrong h after changing volume {}".format(r._h)
    )

    # Change internal radius
    r.r = 14

    # Test if volume returned has proper value, after constraining r
    logAssert(r.volume == 30.0, "Wrong volume after constraining r")

    # Test h was properly modified when constraining r
    logAssert(
        np.isclose(r._h, 30.05, 1e-2), "Wrong h after constraining r {}".format(r._h)
    )


def test_d_can():

    """Tests for diameter of canula (property)"""

    l.debug("Testing d_can")

    r = rea(20, rea.T_CLOSED_ROUND, rea.B_ROUND_IN_O)

    # Test if diameter for canula has valid value
    logAssert(r.d_can == 3, "Wrong d_can returned")

    # Change canula's diameter
    r.d_can = 5

    # Test if canula has proper diameter
    logAssert(r.d_can == 5, "Wrong d_can diameter returned after modif")

    # Check default output is updated when changing d_can
    logAssert(
        r.outputs["default"].height == 5.5,
        "Wrong height for default output after d_can modif",
    )


def test_type_top():

    """Tests for top type (property)"""

    l.debug("Testing top type")

    r = rea(20, rea.T_CLOSED_ROUND, rea.B_ROUND)

    # Test if types of top is properly returned
    logAssert(r.type_top == rea.T_CLOSED_ROUND, "Wrong top type returned")

    r.type_top = rea.T_OPEN

    # Test that changing top type updates h
    logAssert(
        np.isclose(r._h, 38.09, 1e-2), "Wrong h after changing top {}".format(r._h)
    )
    logAssert(
        np.isclose(r._r, 11.77, 1e-2), "Wrong r after changing top {}".format(r._r)
    )


def test_type_bottom():

    """Tests for bottom type (property)"""

    l.debug("Testing bottom type")

    r = rea(20, rea.T_CLOSED_ROUND, rea.B_ROUND)

    # Test if types of top and bottom are properly returned
    logAssert(r.type_bottom == rea.B_ROUND, "Properties: wrong bottom type returned")

    # Change type bottom
    r.type_bottom = rea.B_FLAT_EX_O

    # Create an input to check if it's shifted when changing type_bottom
    r.addInputPercentage("test")

    # Test that changing bottom type updates h
    logAssert(
        np.isclose(r._h, 38.09, 1e-2), "Wrong h after changing top {}".format(r._h)
    )
    logAssert(
        np.isclose(r._r, 11.77, 1e-2), "Wrong r after changing top {}".format(r._r)
    )

    # Check there is a default output after changing type_bottom
    # Check default output is at right height
    logAssert(
        r.outputs["default"].height == 5.5,
        "Wrong h for def output after modif type_bottom",
    )

    # CHange type bottom
    r.type_bottom = rea.B_FLAT_IN_O

    logAssert(
        r.outputs["default"].height == 4.5,
        "Wrong h for def output after 2nd modif type_bottom",
    )

    # Check input 'test' was shifted
    logAssert(
        np.isclose(r.inputs["test"].height, 45.59, 1e-2),
        "Wrong h {} for input 'test' after modif type_bottom".format(
            r.inputs["test"].height
        ),
    )


def test_def_out_angle():

    """Tests for angle for default output (property)"""

    l.debug("Testing default output angle")

    r = rea(20, rea.T_CLOSED_ROUND, rea.B_ROUND_IN_O)

    # Test original angle for def_out_angle
    logAssert(r.def_out_angle == 0, "Properties: wrong value for def_out_angle")

    # Test modification of def_out_angle
    r.def_out_angle = 30

    logAssert(
        np.isclose(np.array(r.def_out_angle), 30, 10e-2),
        "Properties: wrong value for def_out_angle after modif",
    )
    logAssert(
        np.isclose(np.array(r.outputs["default"].angle), 30, 10e-2),
        "Outputs: wrong value for def_out_angle after modif",
    )


def test_types():

    """
    Make sure some important properties have the right type and are returned
    properly
    """

    l.debug("Testing properties")

    r = rea(20, rea.T_CLOSED_ROUND, rea.B_ROUND)

    # Test that property cad is an openscadobject
    logAssert(
        type(r.cad) is solid.objects.translate,
        "Properties: wrong type for reactor.cad: {}".format(type(r.cad)),
    )

    # The CAD code must be a string
    logAssert(type(r.code) is str, "Properties: wrong type for reactor.code")


def test_h_r(all_reactors: list):

    """
    For a 20 mL reactor, check that h and R are right, depending on the
    shape of the reactor
    """

    l.debug("Testing h and r")

    for couple in all_reactors:

        l.debug(
            "Testing h and r for couple Top: {}; Bottom: {}".format(
                couple[0], couple[1]
            )
        )

        r = rea(20, couple[0], couple[1])

        # Closed top, round bottom (without or with inlet)
        if couple[0] in rea.CLOSED_TOPS and couple[1] in rea.ROUND_BOTS:

            logAssert(
                np.isclose(r._h, 36.14, 1e-2), "h is wrong for couple {}".format(couple)
            )
            logAssert(
                np.isclose(r._r, 11.17, 1e-2),
                "R is wrong for couple {}".format(r._r, couple),
            )

        # Treat these cases together: the inside of the reactor has the
        # same shape: closed top and flat bottom, or open top and round bottom
        elif (
            couple[0] in rea.CLOSED_TOPS
            and couple[1] in rea.FLAT_BOTS
            or couple[0] == rea.T_OPEN
            and couple[1] in rea.ROUND_BOTS
        ):

            logAssert(
                np.isclose(r._h, 38.09, 1e-2), "h is wrong for couple {}".format(couple)
            )
            logAssert(
                np.isclose(r._r, 11.77, 1e-2),
                "R is wrong for couple {}".format(r._r, couple),
            )

        # Open top, flat bottom (without or with inlet)
        elif couple[0] in rea.T_OPEN and couple[1] in rea.FLAT_BOTS:

            logAssert(
                np.isclose(r._h, 40.55, 1e-2), "h is wrong for couple {}".format(couple)
            )
            logAssert(
                np.isclose(r._r, 12.53, 1e-2),
                "R is wrong for couple {}".format(r._r, couple),
            )

        # 1/2 inch threaded top, flat bottom
        elif couple[0] == rea.T_2_QUA_INCH and couple[1] in rea.FLAT_BOTS:

            logAssert(
                np.isclose(r._h, 63.67, 1e-2), "h is wrong for couple {}".format(couple)
            )
            logAssert(r._r == 10, "R is wrong for couple {}".format(r._r, couple))

        # 1/2 inch threaded top, round bottom
        elif couple[0] == rea.T_2_QUA_INCH and couple[1] in rea.ROUND_BOTS:

            logAssert(
                np.isclose(r._h, 56.99, 1e-2), "h is wrong for couple {}".format(couple)
            )
            logAssert(r._r == 10, "R is wrong for couple {}".format(r._r, couple))

        # 3/4 inch threaded top, flat bottom
        elif couple[0] == rea.T_3_QUA_INCH and couple[1] in rea.FLAT_BOTS:

            print(r._r)
            print(r._h)

            logAssert(
                np.isclose(r._h, 44.21, 1e-2), "h is wrong for couple {}".format(couple)
            )
            logAssert(r._r == 12, "R is wrong for couple {}".format(r._r, couple))

        # 1/2 inch threaded top, round bottom
        elif couple[0] == rea.T_3_QUA_INCH and couple[1] in rea.ROUND_BOTS:

            logAssert(
                np.isclose(r._h, 36.21, 1e-2), "h is wrong for couple {}".format(couple)
            )
            logAssert(r._r == 12, "R is wrong for couple {} {}".format(r._r, couple))

        # 1 inch threaded top, flat bottom
        elif couple[0] == rea.T_4_QUA_INCH and couple[1] in rea.FLAT_BOTS:

            logAssert(
                np.isclose(r._h, 24.87, 1e-2), "h is wrong for couple {}".format(couple)
            )
            logAssert(r._r == 16, "R is wrong for couple {}".format(r._r, couple))

        # 1/2 inch threaded top, round bottom
        elif couple[0] == rea.T_4_QUA_INCH and couple[1] in rea.ROUND_BOTS:

            logAssert(
                np.isclose(r._h, 14.20, 1e-2), "h is wrong for couple {}".format(couple)
            )
            logAssert(r._r == 16, "R is wrong for couple {}".format(r._r, couple))

        else:
            l.error("output: unrecognized couple: {}; {}".format(couple[0], couple[1]))
            pytest.fail(
                "Output test failed for unrecognized couple: {}; {}".format(
                    couple[0], couple[1]
                )
            )


def test_infos():

    """Make sure the dict 'infos' returns proper values"""

    l.debug("Testing reactor's infos")

    r = rea(20, rea.T_OPEN, rea.B_ROUND)

    t_dict = {
        "r_ex": 14.771641655355797,
        "r": 11.771641655355797,
        "height": 52.865474258855315,
        "height_wo_inlet": 52.865474258855315,
        "height_body": 38.093832603499514,
        "height_bottom": 14.771641655355797,
        "inputs": 0,
        "outputs": 0,
        "walls": 3,
        "type_top": "Open",
        "type_bottom": "Round",
        "volume": 20.0,
        "d_can": 3,
        "align_top_strategy": "expand",
    }

    logAssert(
        t_dict == r.infos,
        "reactor.infos: dicts don't match. Test: {}; \n Target: {}".format(
            r.infos, t_dict
        ),
    )

    # Test that changing walls thickness changes r_ex
    r.walls = 4

    t_dict = {
        "r_ex": 15.771641655355797,
        "r": 11.771641655355797,
        "height": 53.865474258855315,
        "height_wo_inlet": 53.865474258855315,
        "height_body": 38.093832603499514,
        "height_bottom": 15.771641655355797,
        "inputs": 0,
        "outputs": 0,
        "walls": 4,
        "type_top": "Open",
        "type_bottom": "Round",
        "volume": 20.0,
        "d_can": 3,
        "align_top_strategy": "expand",
    }

    logAssert(
        t_dict == r.infos,
        "reactor.infos: dicts don't match. Test: {}; \n Target: {}".format(
            r.infos, t_dict
        ),
    )

    r = rea(20, rea.T_CLOSED_ROUND_O, rea.B_ROUND_IN_O)

    t_dict = {
        "r_ex": 14.168810962106537,
        "r": 11.168810962106537,
        "height": 74.48065342543465,
        "height_wo_inlet": 70.48065342543465,
        "height_body": 36.143031501221586,
        "height_bottom": 20.168810962106537,
        "inputs": 0,
        "outputs": 1,
        "walls": 3,
        "type_top": "Closed round with outlet",
        "type_bottom": "Round with internal outlet",
        "volume": 20.0,
        "d_can": 3,
        "align_top_strategy": "expand",
    }

    logAssert(
        t_dict == r.infos,
        "reactor.infos: dicts don't match. Test: {}; \n Target: {}".format(
            r.infos, t_dict
        ),
    )


def test_describe():

    l.debug("Testing reactor's text description")

    r = rea(60, rea.T_4_QUA_INCH, rea.B_FLAT_EX_O)
    r.addInputPercentage("test")

    comp = "type_top: Threaded cap 1 inch\n"
    comp += "type_bottom: Flat with external outlet\n"
    comp += "volume: 60.0\n"
    comp += "d_can: 3\n"
    comp += "align_top_strategy: expand\n"
    comp += "walls: 3\n"
    comp += "inputs: 1\n"
    comp += "outputs: 1\n"
    comp += "r_ex: 19\n"
    comp += "r: 16\n"
    comp += "height: 99.60387957432594\n"
    comp += "height_wo_inlet: 99.60387957432594\n"
    comp += "height_body: 74.60387957432594\n"
    comp += "height_bottom: 10.0\n"
    comp += "----- Input(s) -----\n"
    comp += "test: {'height': 80.10387957432594, 'diameter': 3, 'angle': 180.0, "
    comp += "'external': False, 'chamber': None}\n"
    comp += "----- Output(s) ----\n"
    comp += "default: {'height': 5.5, 'diameter': 3, 'angle': 0.0, 'external': True, "
    comp += "'chamber': None}"

    logAssert(repr(r.describe()) == repr(comp), "Reactor's description doesn't match")


def test_output():

    """
    Test that reactor has proper output. Just test the height of the output,
    should be enough
    """

    l.debug("Testing reactor's default output")

    # Test only bottoms with inlet
    all_bottoms = [rea.B_ROUND_IN_O, rea.B_ROUND_EX_O, rea.B_FLAT_IN_O, rea.B_FLAT_EX_O]

    for bottom in all_bottoms:

        # Type of top doesn't affect output's height, so only use open top
        r = rea(20, rea.T_OPEN, bottom)

        h = np.array(r.outputs["default"].height)

        if bottom == rea.B_ROUND_EX_O:
            logAssert(
                np.isclose(h, 5.5, 1e-2),
                "Bottom: {}, wrong output h {}".format(bottom, h),
            )

        elif bottom == rea.B_ROUND_IN_O:
            logAssert(
                np.isclose(h, 4.5, 1e-2),
                "Bottom: {}, wrong output h {}".format(bottom, h),
            )

        elif bottom == rea.B_FLAT_EX_O:
            logAssert(
                np.isclose(h, 5.5, 1e-2),
                "Bottom: {}, wrong output h {}".format(bottom, h),
            )

        elif bottom == rea.B_FLAT_IN_O:
            logAssert(
                np.isclose(h, 4.5, 1e-2),
                "Bottom: {}, wrong output h {}".format(bottom, h),
            )

        else:
            l.error("output: unrecognized  bottom".format(bottom))
            pytest.fail("Output test failed for unrecognized bottom: {}".format(bottom))


def test_add_input():

    """Test that inputs are properly added to a reactor"""

    l.debug("Testing add input feature")

    r = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)

    # Check exception is raised when negative value for input height, or
    # if height > 100%
    with pytest.raises(ValueError):
        r.addInputPercentage("new", height_per=-10)
        l.error("Input: no exception raised when height_per < 0")
    with pytest.raises(ValueError):
        r.addInputPercentage("new", height_per=110)
        l.error("Input: no exception raised height_per > 100")

    # Add an input with default params
    r.addInputPercentage("test")

    t_dict = {
        "external": False,
        "angle": 180.0,
        "chamber": None,
        "diameter": 3,
        "height": 45.04822038391718,
    }
    logAssert(r.inputs["test"].infos == t_dict, "Dictionaries do not match")

    # After adding an input, reactor.cad still has the right type
    logAssert(type(r.cad) is solid.objects.translate, "Input: wrong type for reactor")

    # Test the input is a the right place (open top)
    h = np.array(r.inputs["test"].height)
    logAssert(
        np.isclose(h, 45.04, 1e-2),
        "Input: default input at the wrong place (open top) {}".format(h),
    )

    r = rea(20, rea.T_CLOSED_ROUND, rea.B_ROUND)
    r.addInputPercentage("test", 100)

    t_dict = {
        "diameter": 3,
        "angle": 180.0,
        "height": 48.81184246332812,
        "external": False,
        "chamber": None,
    }
    logAssert(r.inputs["test"].infos == t_dict, "Dicitonaries do not match")

    # Test the input is a the right place (closed top)
    h = np.array(r.inputs["test"].height)
    logAssert(
        np.isclose(h, 48.81, 1e-2),
        "Input: default input at the wrong place (closed top) {}".format(h),
    )

    # Change angle of input, and check it's at the right place
    r = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)
    r.addInputPercentage("test", angle=230)

    t_dict = {
        "external": False,
        "diameter": 3,
        "height": 45.04822038391718,
        "angle": 229.99999999999997,
        "chamber": None,
    }
    logAssert(r.inputs["test"].infos == t_dict, "Dictionaries do not match")

    h = np.array(r.inputs["test"].height)
    logAssert(
        np.isclose(h, 45.05, 1e-2),
        "Input: h input with angle 230° at the wrong place {}".format(h),
    )

    # Change height of input, and check it's at the right place
    r = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)
    r.addInputPercentage("test", height_per=0)

    t_dict = {
        "chamber": None,
        "diameter": 3,
        "height": 10.5,
        "angle": 180.0,
        "external": False,
    }
    logAssert(r.inputs["test"].infos == t_dict, "Dictionaries do not match")

    h = np.array(r.inputs["test"].height)
    logAssert(
        np.isclose(h, 10.5, 1e-2),
        "Input: h input with height_per=0 at the wrong place {}".format(h),
    )


def test_add_output():

    """Test that outputs are properly added to a reactor"""

    l.debug("Testing add output feature")

    r = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)

    # Check exception is raised when negative value for output height, or
    # if height > 100%
    with pytest.raises(ValueError):
        r.addOutputPercentage("new", height_per=-10)
        l.error("Output: no exception raised when height_per < 0")
    with pytest.raises(ValueError):
        r.addOutputPercentage("new", height_per=110)
        l.error("Output: no exception raised height_per > 100")

    # Add an output with default params
    r.addOutputPercentage("test")

    t_dict = {
        "diameter": 3,
        "external": False,
        "height": 45.04822038391718,
        "chamber": None,
        "angle": 180.0,
    }
    logAssert(r.outputs["test"].infos == t_dict, "Dictionaries do not match")

    # After adding an input, reactor.cad still has the right type
    logAssert(type(r.cad) is solid.objects.translate, "Output: wrong type for reactor")

    # Test the input is a the right place (open top)
    h = np.array(r.outputs["test"].height)
    logAssert(
        np.isclose(h, 45.05, 1e-2),
        "Output: h default output at the wrong place (open top) {}".format(h),
    )

    r = rea(20, rea.T_CLOSED_ROUND, rea.B_ROUND)
    r.addOutputPercentage("test", 100)

    t_dict = {
        "external": False,
        "diameter": 3,
        "chamber": None,
        "angle": 180.0,
        "height": 48.81184246332812,
    }
    logAssert(r.outputs["test"].infos == t_dict, "Dictionaries do not match")

    # Test the output is a the right place (closed top)
    h = np.array(r.outputs["test"].height)
    logAssert(
        np.isclose(h, 48.81, 1e-2),
        "Output: h default input at the wrong place (closed top) {}".format(h),
    )

    # Change angle of input, and check it's at the right place
    r = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)
    r.addOutputPercentage("test", angle=230)

    t_dict = {
        "angle": 229.99999999999997,
        "diameter": 3,
        "external": False,
        "chamber": None,
        "height": 45.04822038391718,
    }
    logAssert(r.outputs["test"].infos == t_dict, "Dictionaries do not match")

    h = np.array(r.outputs["test"].height)
    logAssert(
        np.isclose(h, 45.05, 1e-2),
        "Output: h input with angle 230° at the wrong place {}".format(h),
    )

    # Change height of input, and check it's at the right place
    r = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)
    r.addOutputPercentage("test", height_per=0)

    t_dict = {
        "chamber": None,
        "external": False,
        "angle": 180.0,
        "diameter": 3,
        "height": 10.5,
    }
    logAssert(r.outputs["test"].infos == t_dict, "Dictionaries do not match")

    h = np.array(r.outputs["test"].height)
    logAssert(
        np.isclose(h, 10.5, 1e-2),
        "Output: h input with height_per=0 at the wrong place {}".format(h),
    )


def test_addLuerTopInlet():

    """
    Test that creating a Luer top inlet works. Don't test createCustomTopInlet
    as this method already calls it
    """

    r = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)

    # Creating Luer top inlet with open top must crash
    with pytest.raises(IncompatibilityError):
        r.addLuerTopInlet("test")
        l.error("Reactor: no exc raised top Luer + not custom top")

    # Create a reactor with custom top, add a Luer top inlet, and check
    # that inlet has proper values
    r = rea(20, rea.T_CUSTOM, rea.B_FLAT_IN_O)
    r.addLuerTopInlet("test")

    # Convert inlet to dict for comparison
    io = r.top_inlets["test"]
    dict_io = {
        "diameter": io.diameter,
        "length": io.length,
        "walls": io.walls,
        "x": io.x,
        "y": io.y,
    }
    compare = {"diameter": 5, "length": 7, "walls": 3, "x": 0, "y": 0}

    logAssert(dict_io == compare, "Luer top inlet not properly created")


def test_getHeightPercentage():

    """Test getting height percentage for InputOutput object works properly"""

    r = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)
    r.addInputPercentage("test", height_per=50)

    height_per = r.getHeightPercentage(r.inputs["test"])

    logAssert(height_per == 50, "getHeightPercentage didn't return proper %")


def test_liftReactor():

    """
    Test the liftReactor method. Make sure infos is properly updated
    after lifting
    """

    l.debug("Testing liftReactor feature")

    r = rea(20, rea.T_CLOSED_ROUND_O, rea.B_ROUND)
    r.liftReactor(20)

    logAssert(
        np.isclose(r.infos["height"], 88.48, 1e-2),
        "reactor.infos: wrong height after lifting",
    )

    # Height without inlet should be the same since T_CLOSED_ROUND has no inlet
    logAssert(
        np.isclose(r.infos["height_wo_inlet"], 84.48, 1e-2),
        "reactor.infos: wrong height wo inlet after lifting",
    )

    logAssert(r.coo["z"] == 20, "reactor.coo: wrong z after lifting")


def test_expandVertically():

    """
    Test the expandVertically method. Make sure volume and height are
    updated after expansion
    """

    l.debug("Testing expandVertically feature")

    r = rea(20, rea.T_CLOSED_ROUND_O, rea.B_ROUND)
    r.expandVertically(20)

    logAssert(
        np.isclose(r.infos["height_wo_inlet"], 84.48, 1e-2),
        "expandVertically: wrong height after expansion",
    )

    logAssert(
        np.isclose(r.volume, 27.84, 1e-2),
        "expandVertically: wrong volume after expansion",
    )


def test_is_connected():

    l.debug("Testing is_connected")

    r = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)

    r.addInputPercentage("new")

    # Simulate connection to input (simply switch a bool)
    r.inputs["new"].connected = True

    logAssert(r.isConnected() is True, "Reactor not connected")


def test_has_internal_input():

    l.debug("Testing if reactor has internal connected input")

    r = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)

    logAssert(r.hasConnectedInternalInput() is False, "Reactor doesn't have any input")

    r.addInputPercentage("new")

    logAssert(
        r.hasConnectedInternalInput() is False,
        "Reactor: pb w internal, unconnected input",
    )

    r.inputs["new"].connected = True

    logAssert(
        r.hasConnectedInternalInput() is True, "Reactor: pb w internal, connected input"
    )

    r.inputs["new"].external = True

    logAssert(r.hasConnectedInternalInput() is False, "Reactor has an external input")


def test_has_internal_connected_output():

    l.debug("Testing if reactor has internal connected output")

    r = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)

    logAssert(
        r.hasInternalConnectedOutput() is False,
        "Reactor has internal but unconnected output",
    )

    r.outputs["default"].connected = True

    logAssert(
        r.hasInternalConnectedOutput() is True, "Reactor has internal connected output"
    )

    r = rea(20, rea.T_OPEN, rea.B_FLAT_EX_O)

    logAssert(r.hasInternalConnectedOutput() is False, "Reactor has external output")


def test_valid_stl(all_reactors: list):

    """
    Generate all possible shapes of reactor, and render the STL files.
    Then test the validity of each STL file:
    - Check if watertight (True if there is non non-manifold surfaces)
    - Check if valid volume
    - Check if valid winding
    """

    l.debug("Testing STLs validity")

    for couple in all_reactors:

        l.debug(
            "Building and rendering couple Top: {}; Bottom: {}".format(
                couple[0], couple[1]
            )
        )

        # Build the reactor
        r = rea(20, couple[0], couple[1])

        # Add inputs to check for non-manifolds
        r.addInputPercentage("i1", 0)
        r.addInputPercentage("i2", 100)

        # Forge the name of the file
        name = "{} {}".format(couple[0], couple[1])
        name = name.replace(" ", "_")
        name = name.replace("/", "_")
        r.renderToFile(name + ".scad")

        # Render the reactor into a stl file
        cmd = sp.Popen(
            ["openscad", "-o", name + ".stl", name + ".scad"],
            stdout=sp.PIPE,
            stderr=sp.PIPE,
        )

        # Wait until rendering is finished
        out, err = cmd.communicate()

        # Check if rendering succeeded, crash the test otherwise
        if cmd.returncode != 0:
            l.error(
                "Failed to render STL for couple Top {}; Bottom: {}".format(
                    couple[0], couple[1]
                )
            )
            pytest.fail(
                "Failed to render STL for couple Top: {}; Bottom: {}".format(
                    couple[0], couple[1]
                )
            )

        # Generate the mesh from STL file
        mesh = trimesh.load_mesh(name + ".stl")

        logAssert(
            mesh.is_watertight,
            "Couple Top: {}; Bottom: {} has non-manifold surfaces".format(
                couple[0], couple[1]
            ),
        )

        logAssert(
            mesh.is_volume,
            "Couple Top: {}; Bottom: {} has a negative volume".format(
                couple[0], couple[1]
            ),
        )

        logAssert(
            mesh.is_winding_consistent,
            "Couple Top: {}; Bottom: {} has unconsistent winding".format(
                couple[0], couple[1]
            ),
        )

        # Remove files to not polute dir
        os.remove(name + ".scad")
        os.remove(name + ".stl")


if __name__ == "__main__":
    pass



#!/usr/bin/python
# coding: utf-8

from vispy.visuals.shaders import Function, Varying
from vispy.color import colormap, Color
import os
from vispy import app, geometry, gloo
import vispy.scene
from vispy.scene import visuals
from vispy.app import use_app
from chemscad.log import MyLog
from typing import Tuple, Union, Dict
from vispy.visuals.filters import Alpha
import vispy.io
from vispy.gloo.util import _screenshot
import numpy as np
from chemscad.app_constants import colours_module_type_short
from glob import glob

use_app("PyQt5")


class STLViewer(vispy.scene.SceneCanvas):

    """"
    Class to view a STL file. Camera is embedded, so rotations/translations are
    possible
    https://github.com/vispy/vispy/issues/1425
    https://github.com/vispy/vispy/issues/1479
    """

    # Initial direction for light, for all meshes
    INI_LIGHT_DIR = (0, -1, 0)

    # Alpha value when a module is transparent
    TRANSPARENT_ALPHA_CST = 0.4

    def __init__(
        self,
        size: Tuple[int, int] = (800, 500),
        watch_dir: str = ".",
        logger: Union[None, MyLog] = None,
    ) -> None:

        """
        size is the size in pixels (x, y). watch_dir is the path to watch
        for new STL files
        """

        vispy.scene.SceneCanvas.__init__(
            self, size=size, bgcolor="black", #keys="interactive",
        )

        # Unfreeze the class, so adding attributes is possible
        self.unfreeze()

        if logger is None:
            self.l = MyLog("activity.log")
        else:
            self.l = logger

        self.watch_dir = watch_dir

        # Create a dict to store several meshes
        # Meshes are stored with their names so they can be reloaded/reviewed
        # namely
        self.meshes: Dict[str, visuals.Mesh] = dict()

        # Add an 'arcball' camera
        self.view = self.central_widget.add_view()
        self.view.camera = "arcball"

        # Unzoom by default. Otherwise, too close to object at creation
        self.view.camera.scale_factor = 220

        # Store if perspective is enabled. Used by main window
        self.perspective = False

        # Translucent state for Meshes
        self.translucent = False



        # Inverse transform of the initial light dir. Needed to undo the
        # initial orientation of the camera
        direction = self.view.camera.transform.imap(self.INI_LIGHT_DIR)[:3]

        # Compensate for shift (translations) of the camera. This way the
        # light direction only depends of the local orientation of the camera,
        # not on the point it is looking at. For more details, see issue
        # vispy's issue #1479
        self.ini_light_dir = np.concatenate((direction, [0]))

        # Freeze the class again. Might not be necessary
        self.freeze()

    def togglePerspective(self):

        """Toggle the perspective view"""

        self.perspective = not self.perspective

        if self.perspective:
            self.view.camera.fov = 50
        else:
            self.view.camera.fov = 0

    def toggleTranslucent(self):

        """Toggle the translucent state of all the meshes. Called X-ray in GUI"""

        self.translucent = not self.translucent

        for mesh in self.meshes.values():

            # Change the alpha value of the mesh's alpha filter
            # Disable depht_test when translucent, better visual
            # https://github.com/vispy/vispy/issues/1482
            if self.translucent:
                self.view.camera.fov = 50 #HPD - fix x-ray view overlap issue by adding FOV
                if mesh.filter.alpha != 0:
                    mesh.filter.alpha = self.TRANSPARENT_ALPHA_CST
                    mesh.set_gl_state(depth_test=False)
            else:
                self.view.camera.fov = 0 #HPD - as above for FOV
                if mesh.filter.alpha != 0:
                    mesh.filter.alpha = 1
                    mesh.set_gl_state(depth_test=True)

        # Update the scene immediately
        self.view.scene.update()

    def viewMesh(self, file_name: str):

        """
        Will load the file file_name in the STL viewer. If the mesh didn't
        exist before, it will be created. Else, it will be updated
        """

        # file_name is passed wo extension, so add .stl
        file_name = file_name + ".stl"

        file_to_view = os.path.join(self.watch_dir, file_name)

        self.l.debug(f"Trying to load STL file {file_to_view}")

        try:
            # Load the stl file. Works with binary or ascii stl files
            raw_mesh = vispy.io.stl.load_stl( open(file_to_view, 'r') )
            mdata = geometry.MeshData(raw_mesh['vertices'], raw_mesh['faces'])


        except Exception as e:
            self.l.error(
                f"{e}: Failed to load STL {file_to_view} in Canvas", exc_info=True
            )
            return

        # Try to update the mesh, if it doesn't exist, create it
        try:
            self.meshes[file_name].set_data(meshdata=mdata)
            self.l.debug(f"Updating STL {file_name}")
        except KeyError:
            self.l.debug(f"Mesh for {file_name} doesn't exist yet, creating")

            module_type_short = file_name.split(".scad")[0]
            module_type_short = "".join(
                [i for i in module_type_short if not i.isdigit()]
            )

            # Create a Mesh visual, with the view as parent
            mesh = visuals.Mesh(
                meshdata=mdata,
                shading="flat",
                parent=self.view.scene,
                color=colours_module_type_short[module_type_short],
            )

            # Set up the shininess (0 means mat)
            # set up the initial light direction for the mesh
            # Create an alpha filter. alpha=1
            mesh.light_dir = self.INI_LIGHT_DIR
            mesh.shininess = 0
            translucent_filter = Alpha()

            self.meshes[file_name] = mesh

            # Unfreeze the mesh, and make the alpha filter one of
            # its attributes
            self.meshes[file_name].unfreeze()
            self.meshes[file_name].filter = translucent_filter
            self.meshes[file_name].attach(translucent_filter)
            self.meshes[file_name].freeze()

            # Disable depth_test when translucent, better visual
            if self.translucent:
                mesh.set_gl_state(depth_test=False)

    def deleteMesh(self, file_name: str):
        """
        Deletes a mesh from meshes dict. This is invoked when a user
        deletes one of the modules
        """

        # First we will delete the mesh of the acutal object to delete
        dict_key = file_name+".scad.stl"

        try:
            mesh = self.meshes[dict_key]
            mesh.parent = None # based on vispy this is what actually deletes it
            del mesh
            del self.meshes[dict_key]
            self.l.debug(f"Deleting mesh {dict_key}")
        except KeyError:
            self.l.debug(f"Mesh for {dict_key} doesn't exist yet. Cannot delete")

        # Now, because JP doesn't want to show disconnected objects, his code
        # automatically deletes all the STLs and SCADs of disconnected objects.
        # Meaning that if the object to delete would create disconnected objects
        # their files will be deleted and we should also remove them from the mesh.
        # Therefore we will the files left in the render folder, and delete
        # the meshes that don't have a file in render.
        # Disconnected objects that were not user-deleted are still part of
        # the assembly, and showed in the dashboard, but not show in the
        # 3D visualization

        # first obtain the stl files still in the watch folder
        stls = [os.path.basename(stl) for stl in
            glob(self.watch_dir+'/*.stl')]

        # now delete the meshes without an associated stl file
        for dict_key in list(self.meshes):

            # if the mesh doesnt have a file in renders, delete it
            if dict_key not in stls:
                try:
                    mesh = self.meshes[dict_key]
                    mesh.parent = None
                    del mesh
                    del self.meshes[dict_key]
                    self.l.debug(f"Deleting mesh {dict_key}")
                except KeyError:
                    self.l.debug(f"Mesh for {dict_key} doesn't exist yet. Cannot delete")

        self.view.scene.update()

    def toggleVisibility(self, module_id: str):

        """
        If switch_state == True make Mesh associated with file name visible,
        otherwise make the Mesh invisible.
        """

        # Forge the file_name from the module_id
        file_name = module_id + ".scad.stl"
        mesh = self.meshes[file_name]

        # Toggle visibility by changing alpha
        if mesh.filter.alpha != 0:
            mesh.filter.alpha = 0
            mesh.set_gl_state(depth_test=False)
        else:
            if self.translucent:
                mesh.filter.alpha = self.TRANSPARENT_ALPHA_CST
                mesh.set_gl_state(depth_test=False)
            else:
                mesh.filter.alpha = 1
                mesh.set_gl_state(depth_test=True)

        self.view.scene.update()

    def screenshot(self, path: str):

        """
        Render the current canvas to an image and save the image
        https://stackoverflow.com/questions/28732937/how-to-export-vispy-result-as-png-image
        """

        img = _screenshot()
        vispy.io.write_png(path, img)

    def on_mouse_move(self, event):

        """
        Called when mouse is clicked and is moving. Sub-classed to make
        the light follow the camera
        """

        transform = self.view.camera.transform

        for name, mesh in self.meshes.items():

            # Forward transform on light dir, light follows camera
            mesh.light_dir = transform.map(self.ini_light_dir)[:3]

    # def on_mouse_release(self, event):

    # self.screenshot()

    # """Attempt to change the light of the mesh when mouse is released"""

    # for mesh in self.meshes.values():
    # Doesn't work bc position of light_dir follows the mesh
    # mesh.light_dir = (0, -50, 50)

    # self.view.scene.update()


if __name__ == "__main__":
    win = STLViewer()
    win.viewMesh("R1.scad")
    win.show()

    import sys

    if sys.flags.interactive != 1:
        app.run()
        print(app)



#!/usr/bin/python
# coding: utf-8

import sys
from PyQt5 import QtGui, QtCore, QtWidgets
from typing import Union, List, Dict, Optional
from ccad import ReactorCAD
from ccad.object import ObjectCAD

from chemscad.log import MyLog


class Settings(QtWidgets.QDialog):

    """
    Display a dialog box. The user can enter his settings here.
    """

    def __init__(self, parent=None):

        super(Settings, self).__init__(parent)

        self.setModal(True)

        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

        self.parent = parent

        if parent is None:
            self.l = MyLog("activity.log")
            self.test = True
        else:
            self.l = self.parent.l
            self.options = self.parent.options
            self.test = False

        self.initUI()
        self.defineSlots()
        self.restoreParams()

    def defineSlots(self):

        """Establish the slots"""

        # User finished selecting tops to align
        self.ok_button.clicked.connect(self.readForm)

    def restoreParams(self):

        """
        Restore the user's preferences
        """

        if self.test:
            return

        # Restore turn_x and turn_y from Assembly
        self.spin_x.setValue(self.parent.ass.turn_x)
        self.spin_y.setValue(self.parent.ass.turn_y)

    def readForm(self):

        """
        Read the preferences
        """

        if self.test:
            return

        # Read turn_x and turn_y for assembly
        self.parent.ass.turn_x = self.spin_x.value()
        self.parent.ass.turn_y = self.spin_y.value()

        # General refresh from main window
        self.parent.refresh()

        self.close()

    def initUI(self):

        """Handles the display"""

        self.setWindowTitle("Assembly angle settings")
        self.setMinimumSize(400, 100)

        fbox = QtWidgets.QFormLayout()

        self.spin_x = QtWidgets.QSpinBox(self)
        self.spin_x.setMinimum(0)
        self.spin_x.setValue(2)

        self.spin_y = QtWidgets.QSpinBox(self)
        self.spin_y.setMinimum(0)
        self.spin_y.setValue(2)

        fbox.addRow("X turn:", self.spin_x)
        fbox.addRow("Y turn:", self.spin_y)

        self.ok_button = QtWidgets.QPushButton("OK", self)

        # ----------------- ASSEMBLING ----------------------------------------

        self.vbox_global = QtWidgets.QVBoxLayout(self)
        self.vbox_global.addLayout(fbox)
        self.vbox_global.addWidget(self.ok_button)

        self.setLayout(self.vbox_global)
        self.show()
    


if __name__ == "__main__":

    app = QtWidgets.QApplication(sys.argv)
    obj = Settings()
    sys.exit(app.exec_())



#!/usr/bin/python
# coding: utf-8


import sys
import os
from typing import Union
from PyQt5 import QtGui, QtCore, QtWidgets

from ccad import DoubleFilterReactorCAD as df_rea
from add_io_ffr import AddIODialogFFR
from add_top_io_ffr import AddTopIODialogFFR
from ccad.exceptions import *
import ccad.constants as cst


class ChemModule(QtWidgets.QDialog):

    """
    Double filter reactor module. Simple cartridge where a reaction
    could happen
    """

    # Define str for type of I/O
    SIDE_INPUT = "Side input"
    SIDE_OUTPUT = "Side output"
    DEF_OUTPUT = "Default output"
    TOP_LUER = "Luer top inlet"
    TOP_CUSTOM = "Custom top inlet"

    def __init__(self, parent=None, cad_obj: Union[None, df_rea] = None):

        super(ChemModule, self).__init__(parent)

        self.setModal(True)

        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

        self.parent = parent

        self.cad_obj = cad_obj

        if parent is None:
            sys.path.append(
                os.path.dirname(
                    os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
                )
            )
            from log import MyLog

            self.l = MyLog("activity.log")

            # # Dummy file for saving if testing
            # self.options = QtCore.QSettings("debug/options.ini",
            # QtCore.QSettings.IniFormat)

            self.test = True
        else:
            self.l = self.parent.l
            # self.options = self.parent.options
            self.test = False

        self.path = os.path.dirname(os.path.abspath(__file__))
        self.l.debug(f"Module path: {self.path}")

        self.initUI()
        self.defineSlots()

        # Store inputs
        self.dict_inputs = dict()

        # Store outputs
        self.dict_outputs = dict()

        # Store top inlets
        self.dict_top_inlets = dict()

        # Dict to store module's features
        self.params = dict()
        self.restoreParams()

    def defineSlots(self):

        """Establish the slots"""

        # Bottom combo box changed
        self.combo_bottom.currentIndexChanged.connect(self.bottomTypeChanged)

        # Align top strategy changed
        self.combo_align_top_strategy.currentIndexChanged.connect(
            self.alignTopStrategyChanged
        )

        # To close the window and build the module
        self.ok_button.clicked.connect(self.buildModule)

        # To close the window and delete the module
        self.del_button.clicked.connect(self.deleteModule)
        self.del_button.setShortcut("Del")

        # Open dialog box to add I/O
        self.button_add_io.clicked.connect(self.openTypeIODialog)

        # Open dialog box to modify I/O from list of inputs
        self.table_io.itemDoubleClicked.connect(self.openIODialog)

        # Delete an I/O when button clicked
        self.button_delete_io.clicked.connect(self.deleteIO)

    def bottomTypeChanged(self):

        """
        Slot called when combo box for bottom changes. Will disable the
        spin box for diameter of internal pipe when bottom has no pipe
        """

        bottom_type = self.combo_bottom.currentText()

        if bottom_type in df_rea.BOTTOMS_PIPE:
            self.spin_d_can.setEnabled(True)
        else:
            self.spin_d_can.setEnabled(False)

    def alignTopStrategyChanged(self):

        """
        Called when strategy to align top changes. If strategy is lift,
        disable the expand strategy combo box
        """

        strategy = self.combo_align_top_strategy.currentText()

        if strategy == "Lift reactor":
            self.combo_expand_strategy.setEnabled(False)
        elif strategy == "Expand body":
            self.combo_expand_strategy.setEnabled(True)

    def deleteIO(self):

        """Slot: delete an input from the reactor"""

        # Get row for selected I/O
        try:
            row = self.table_io.selectionModel().selectedRows()[0].row()
        except IndexError:
            self.l.debug("No I/O selected, can't delete")
            return

        # Get name and type for I/O
        name = self.table_io.item(row, 0).text()
        type_io = self.table_io.item(row, 1).text()

        if type_io == self.DEF_OUTPUT:
            # We should never be here since selection is disabled for def
            # output, but just in case, return
            return

        elif type_io == self.SIDE_INPUT:
            del self.dict_inputs[name]
            # Delete I/O from CAD object
            if self.cad_obj is not None:
                del self.cad_obj.inputs[name]

        elif type_io == self.SIDE_OUTPUT:
            del self.dict_outputs[name]
            if self.cad_obj is not None:
                del self.cad_obj.outputs[name]

        else:
            del self.dict_top_inlets[name]
            if self.cad_obj is not None:
                del self.cad_obj.top_inlets[name]

        self.l.debug(f"Double Filter reactor, deleting I/O {name}")

        # Remove I/O line from table
        self.table_io.removeRow(row)

    def openTypeIODialog(self):

        """
        Create a small dialog box to let the user choose what type of I/O
        he wants to create: side I/O, or top I/O
        """

        dial = QtWidgets.QDialog(self)

        label_choice = QtWidgets.QLabel("What would you like to create:")

        button_side_io = QtWidgets.QPushButton("A side I/O")
        button_top_io = QtWidgets.QPushButton("A top I/O")

        button_side_io.clicked.connect(dial.accept)
        button_side_io.clicked.connect(lambda: AddIODialogFFR(self))

        button_top_io.clicked.connect(dial.accept)
        button_top_io.clicked.connect(lambda: AddTopIODialogFFR(self))

        hbox_dial = QtWidgets.QHBoxLayout()
        hbox_dial.addWidget(button_side_io)
        hbox_dial.addWidget(button_top_io)

        vbox_dial = QtWidgets.QVBoxLayout()
        vbox_dial.addWidget(label_choice, alignment=QtCore.Qt.AlignHCenter)
        vbox_dial.addLayout(hbox_dial)

        dial.setLayout(vbox_dial)

        dial.show()

    def openIODialog(self, item: QtWidgets.QTableWidgetItem):

        """
        Slot: open I/O dialog box when an I/O is double-clicked in the list
        of I/O
        """

        self.l.debug("Double Filter Reactor, opening I/O dialog box")

        # Get row for double-clicked I/O
        row = item.row()

        # Get name and type for I/O
        name = self.table_io.item(row, 0).text()
        type_io = self.table_io.item(row, 1).text()

        # Create a dict with name and type of I/O
        params = {"name": name, "type_io": type_io}

        # Get parameters for I/O from dicts
        if type_io == self.SIDE_INPUT:
            # Merge simple dict params with full dict for I/O
            params = {**params, **self.dict_inputs[name]}
            AddIODialogFFR(self, params)
        elif type_io == self.SIDE_OUTPUT:
            params = {**params, **self.dict_outputs[name]}
            AddIODialogFFR(self, params)
        else:
            params = {**params, **self.dict_top_inlets[name]}
            AddTopIODialogFFR(self, params)

    def restoreParams(self):

        """
        If a CAD object is provided, we are updating the object.
        Restore the parameters previously entered to the display
        """

        # If no cad_obj, we're creating a module, nothing to restore
        if self.cad_obj is None:
            return

        # Get the values of the object from cad object
        self.params["volume_top"] = self.cad_obj.volume_top
        self.params["volume_bottom"] = self.cad_obj.volume_bottom
        self.params["type_top"] = self.cad_obj.type_top
        self.params["type_bottom"] = self.cad_obj.type_bottom
        self.params["h_filter"] = self.cad_obj.h_ffilter
        self.params["d_filter"] = self.cad_obj.d_ffilter

        # Advanced settings
        self.params["d_can"] = self.cad_obj.d_can

        # Set values for graphical elements
        self.spin_vol_top.setValue(self.params["volume_top"])
        self.spin_vol_bottom.setValue(self.params["volume_bottom"])
        self.combo_top.setCurrentText(self.params["type_top"])
        self.combo_bottom.setCurrentText(self.params["type_bottom"])
        self.spin_h_filter.setValue(self.params["h_filter"])
        self.spin_d_filter.setValue(self.params["d_filter"])
        self.spin_d_can.setValue(self.params["d_can"])

        # Restore the align top strategy
        align_top_strategy = self.cad_obj.align_top_strategy
        if align_top_strategy == "expand":
            self.combo_align_top_strategy.setCurrentText("Expand body")
        elif align_top_strategy == "lift":
            self.combo_align_top_strategy.setCurrentText("Lift reactor")

        # Restore the expand top strategy
        expand_strategy = self.cad_obj.expand_strategy
        if expand_strategy == "expand_top":
            self.combo_expand_strategy.setCurrentText("Expand top chamber")
        elif expand_strategy == "expand_bottom":
            self.combo_expand_strategy.setCurrentText("Expand bottom chamber")

        # Restore the align filter strategy
        align_filter_strategy = self.cad_obj.align_filter_strategy
        if align_filter_strategy == "move_filter":
            self.combo_align_filter_strategy.setCurrentText("Move filter")
        elif align_filter_strategy == "lift":
            self.combo_align_filter_strategy.setCurrentText("Lift reactor")

        self.restoreIO()

        self.l.debug(f"Double filter reactor, restored reactor: {self.params}")

    def restoreIO(self):

        """Restore the I/O of the cad object into the dialog box"""

        # Restore all inputs for object
        for name, in_io in self.cad_obj.inputs.items():
            height_per = self.cad_obj.getHeightPercentage(in_io)
            chamber = in_io.chamber
            angle = in_io.angle
            diameter = in_io.diameter
            external = in_io.external
            connected = in_io.connected

            params = {
                "height_per": height_per,
                "angle": angle,
                "diameter": diameter,
                "external": external,
                "connected": connected,
                "chamber": chamber,
            }

            self.dict_inputs[name] = params

            # Add I/O to the table
            self._addNewRow(name, self.SIDE_INPUT, connected)

        # Restore all outputs for object
        for name, out_io in self.cad_obj.outputs.items():
            height_per = self.cad_obj.getHeightPercentage(out_io)
            chamber = out_io.chamber
            angle = out_io.angle
            diameter = out_io.diameter
            external = out_io.external
            connected = out_io.connected

            params = {
                "height_per": height_per,
                "angle": angle,
                "diameter": diameter,
                "external": external,
                "connected": connected,
                "chamber": chamber,
            }

            self.dict_outputs[name] = params

            # Add I/O to the table
            if name == "default":
                # Different type of output if default
                self._addNewRow(name, self.DEF_OUTPUT, connected)
            else:
                self._addNewRow(name, self.SIDE_OUTPUT, connected)

        # Restore all top inlets for object
        for name, top_io in self.cad_obj.top_inlets.items():
            diameter = top_io.diameter
            length = top_io.length
            walls = top_io.walls
            type_io = top_io.type_io
            connected = False

            params = {
                "diameter": diameter,
                "length": length,
                "walls": walls,
                "type_io": type_io,
                "connected": connected,
            }

            self.dict_top_inlets[name] = params

            # Add I/O to the table
            self._addNewRow(name, type_io, connected)

        self.l.debug("Double filter reactor, restored I/O")

    def readForm(self):

        """
        Read the form, and get the parameters of the module. Put them in a dict
        """

        self.params["volume_top"] = self.spin_vol_top.value()
        self.params["volume_bottom"] = self.spin_vol_bottom.value()
        self.params["type_top"] = self.combo_top.currentText()
        self.params["type_bottom"] = self.combo_bottom.currentText()
        self.params["h_filter"] = self.spin_h_filter.value()
        self.params["d_filter"] = self.spin_d_filter.value()

        # Advanced settings
        self.params["d_can"] = self.spin_d_can.value()

        # Get align top strategy and transform to usable string value
        strategy = self.combo_align_top_strategy.currentText()
        if strategy == "Expand body":
            strategy = "expand"
        elif strategy == "Lift reactor":
            strategy = "lift"
        self.params["align_top_strategy"] = strategy

        # Get align top strategy and transform to usable string value
        align_top_strategy = self.combo_align_top_strategy.currentText()
        if align_top_strategy == "Expand body":
            align_top_strategy = "expand"
        elif align_top_strategy == "Lift reactor":
            align_top_strategy = "lift"
        self.params["align_top_strategy"] = align_top_strategy

        # Get expand strategy and transform to usable string value
        expand_strategy = self.combo_expand_strategy.currentText()
        if expand_strategy == "Expand top chamber":
            expand_strategy = "expand_top"
        elif expand_strategy == "Expand bottom chamber":
            expand_strategy = "expand_bottom"
        self.params["expand_strategy"] = expand_strategy

        # Get align filter strategy and transform to usable string value
        align_filter_strategy = self.combo_align_filter_strategy.currentText()
        if align_filter_strategy == "Move filter":
            align_filter_strategy = "move_filter"
        elif align_filter_strategy == "Lift reactor":
            align_filter_strategy = "lift"
        self.params["align_filter_strategy"] = align_filter_strategy

        self.l.debug(f"Double filter reactor, read form {self.params}")

    def addInput(self, params: dict):

        """The AddIODialogFFR class will call this method"""

        self.l.debug(f"Double filter reactor, adding input: {params}")

        name = params["name"]
        del params["name"]

        # If input not already in dict_inputs, add it to the qlist
        if name not in self.dict_inputs:
            self._addNewRow(name, params["type_io"])
            self.l.debug(f"module_double_reactor: addInput {params}")
        else:
            self.l.debug(f"addInput, updating input {name} with {params}")

        self.dict_inputs[name] = params

    def addOutput(self, params: dict):

        """The AddIODialogFFR class will call this method"""

        self.l.debug(f"Double filter reactor, adding output: {params}")

        name = params["name"]
        del params["name"]

        # If input not already in dict_outputs, add it to the qlist
        if name not in self.dict_outputs:
            self._addNewRow(name, params["type_io"])
            self.l.debug(f"module_double_filter_reactor: addOutput {params}")
        else:
            self.l.debug(f"addOutput, updating output {name} with {params}")

        self.dict_outputs[name] = params

    def addTopInlet(self, params: dict):

        """
        The AddTopIODialogFFR class will call this method. Create a top inlet
        """

        self.l.debug(f"Double Filter Reactor, adding top inlet: {params}")

        name = params["name"]
        del params["name"]

        # If input not already in dict_top_inlets, add it to the table
        if name not in self.dict_top_inlets:
            self._addNewRow(name, params["type_io"])
            self.l.debug(f"module_double_filter_reactor: top inlet {params}")
        else:
            self.l.debug(f"addTopInlet, updating inlet {name} with {params}")

        self.dict_top_inlets[name] = params

    def _addNewRow(self, name: str, type_io: str, connected: bool = False):

        """Add a new row in the I/O table"""

        self.l.debug(f"Adding new row {name}, {type_io}")

        # Count nbr of rows
        row_count = self.table_io.rowCount()

        # Insert row at the end
        self.table_io.insertRow(row_count)

        # Create name and type of I/O items to insert in the table
        item_name = QtWidgets.QTableWidgetItem(name)

        # TopInlet objects have an attribute 'type_io'. InputOutput objects
        # don't, their type is determined based on which dict they are in
        # (inputs or outputs)
        if type_io == "custom":
            item_type_io = QtWidgets.QTableWidgetItem("Custom top inlet")
        elif type_io == "luer":
            item_type_io = QtWidgets.QTableWidgetItem("Luer top inlet")
        else:
            item_type_io = QtWidgets.QTableWidgetItem(type_io)

        # Create an item depending on the connection state of the I/O
        if connected:
            item_connected = QtWidgets.QTableWidgetItem("Connected")
        else:
            item_connected = QtWidgets.QTableWidgetItem("Not connected")

        # Disable selection for default output; can't do anything on it
        # from I/O tab
        if type_io == self.DEF_OUTPUT:
            item_name.setFlags(QtCore.Qt.NoItemFlags)
            item_type_io.setFlags(QtCore.Qt.NoItemFlags)
            item_connected.setFlags(QtCore.Qt.NoItemFlags)

        # Insert the 2 items in the table
        self.table_io.setItem(row_count, 0, item_name)
        self.table_io.setItem(row_count, 1, item_type_io)
        self.table_io.setItem(row_count, 2, item_connected)

    def buildInputs(self, cad_obj):

        """Build an input for the CAD object"""

        self.l.debug("Double filter reactor, building inputs")

        for name, in_io in self.dict_inputs.items():

            if in_io["chamber"] == "top":
                cad_obj.addInputPercentageTop(
                    name,
                    in_io["height_per"],
                    in_io["angle"],
                    in_io["diameter"],
                    in_io["external"],
                )

            if in_io["chamber"] == "bottom":
                cad_obj.addInputPercentageBottom(
                    name,
                    in_io["height_per"],
                    in_io["angle"],
                    in_io["diameter"],
                    in_io["external"],
                )

    def buildOutputs(self, cad_obj):

        """Build an ouput for the CAD object"""

        self.l.debug("Double filter reactor, building outputs")

        for name, out_io in self.dict_outputs.items():
            if name == "default":
                continue

            if out_io["chamber"] == "top":
                cad_obj.addOutputPercentageTop(
                    name,
                    out_io["height_per"],
                    out_io["angle"],
                    out_io["diameter"],
                    out_io["external"],
                )

            if out_io["chamber"] == "bottom":
                cad_obj.addOutputPercentageBottom(
                    name,
                    out_io["height_per"],
                    out_io["angle"],
                    out_io["diameter"],
                    out_io["external"],
                )

    def buildTopInlets(self, cad_obj):

        """
        Build the top inlets. Also try to place them. For now, only
        auto-placement is supported
        """

        self.l.debug("Double Filter Reactor, building top inlets")

        for name, io in self.dict_top_inlets.items():

            try:
                cad_obj.addCustomTopInlet(
                    name, io["diameter"], io["length"], io["walls"], io["type_io"]
                )
            except IncompatibilityError:
                self.l.debug("can't build top inlet: no custom top")
                return

        # TODO: handle manual placement (later)
        try:
            cad_obj.autoPlaceTopInlets()
        except ConstraintError:
            self.l.debug("Can't auto-place top inlets, collision ?")
            mes = """
            Impossible to auto-place the top inlets. Collision ?
            """

            QtWidgets.QMessageBox.critical(
                self, "Auto-placement error", mes, QtWidgets.QMessageBox.Ok
            )

            return

    def buildModule(self):

        """
        Called when user clicks the OK button. Will read the form, build the 3D
        model of the module, and tell the parent to display the module
        """

        # TODO: generate an image of the module, and build the CAD object

        self.readForm()

        if not self.test:
            # If no CAD object provided, we're creating a new object
            if self.cad_obj is None:
                self._buildNewModule()
            else:
                self._updateModule()

        # Close the settings window and free the memory
        self.close()

    def deleteModule(self):

        """
        Called when user clicks the delete button. 
        It will delete the selected module
        """

        self.l.debug("Reactor, deleteModule")
        self.parent.deleteModule(self.cad_obj)

        # Close the settings window and free the memory
        self.close()

    def _buildNewModule(self) -> bool:

        """
        Internal method called when building the object for the first time
        """

        self.l.debug("Double filter reactor, _buildNewModule")

        try:
            reactor = self._buildCADObject()
        except Exception as e:
            mes = f"Impossible to create double filter reactor: {e}"
            self.l.error(mes, exc_info=True)
            QtWidgets.QMessageBox.critical(
                self, "Creation error", mes, QtWidgets.QMessageBox.Ok
            )
            return False

        # Try to build I/Os
        try:
            self.buildInputs(reactor)
            self.buildOutputs(reactor)
            self.buildTopInlets(reactor)
        except Exception as e:
            mes = f"Impossible to create I/Os: {e}"
            self.l.error(mes, exc_info=True)
            QtWidgets.QMessageBox.critical(
                self, "I/Os error", mes, QtWidgets.QMessageBox.Ok
            )
            return False

        self.parent.buildModule(reactor)

        return False

    def _updateModule(self) -> bool:

        """
        Internal method called when updating the object with new form values
        """

        self.l.debug("Double fitler reactor, _updateModule")

        # Try to build a temporary reactor with the new parameters
        # If it fails, exit immediately
        try:
            _ = self._buildCADObject()
        except Exception as e:
            mes = f"Impossible to update reactor: {e}"
            self.l.error(mes, exc_info=True)
            QtWidgets.QMessageBox.critical(
                self, "Update error", mes, QtWidgets.QMessageBox.Ok
            )
            return False

        vol_top = self.params["volume_top"]
        vol_bot = self.params["volume_bottom"]
        self.cad_obj.volume_bottom = vol_bot
        self.cad_obj.volume_top = vol_top
        self.cad_obj.volume = vol_top + vol_bot
        self.cad_obj.type_top = self.params["type_top"]
        self.cad_obj.type_bottom = self.params["type_bottom"]
        self.cad_obj.h_ffilter = self.params["h_filter"]
        self.cad_obj.d_ffilter = self.params["d_filter"]
        self.cad_obj.h_bfilter = self.params["h_filter"]
        self.cad_obj.d_bfilter = self.params["d_filter"]
        self.cad_obj.d_can = self.params["d_can"]
        self.cad_obj.align_top_strategy = self.params["align_top_strategy"]
        self.cad_obj.expand_strategy = self.params["expand_strategy"]
        self.cad_obj.align_filter_strategy = self.params["align_filter_strategy"]

        # Try to build I/Os
        try:
            self.buildInputs(self.cad_obj)
            self.buildOutputs(self.cad_obj)
            self.buildTopInlets(self.cad_obj)
        except Exception as e:
            mes = f"Impossible to create I/Os: {e}"
            self.l.error(mes, exc_info=True)
            QtWidgets.QMessageBox.critical(
                self, "I/Os error", mes, QtWidgets.QMessageBox.Ok
            )
            return False

        try:
            self.parent.refresh()
        except Exception as e:
            mes = f"Impossible to refresh double filter reactor: {e}"
            self.l.error(mes, exc_info=True)
            QtWidgets.QMessageBox.critical(
                self, "Refresh error", mes, QtWidgets.QMessageBox.Ok
            )
            return False

        return True

    def _buildCADObject(self) -> df_rea:

        """
        Use the parameters read from the form to build the CAD object. Does
        nothing else, no GUI task here
        """

        volume_top = self.params["volume_top"]
        volume_bottom = self.params["volume_bottom"]
        type_top = self.params["type_top"]
        type_bottom = self.params["type_bottom"]
        d_filter = self.params["d_filter"]
        h_filter = self.params["h_filter"]

        # Advanced settings
        d_can = self.params["d_can"]
        align_top_strategy = self.params["align_top_strategy"]
        expand_strategy = self.params["expand_strategy"]
        align_filter_strategy = self.params["align_filter_strategy"]

        reactor = df_rea(
            volume_bottom,
            volume_top,
            type_top,
            type_bottom,
            d_filter,
            h_filter,
            d_can=d_can,
            align_top_strategy=align_top_strategy,
            expand_strategy=expand_strategy,
            align_filter_strategy=align_filter_strategy,
        )

        return reactor

    def initUI(self):

        """Handles the display"""

        if self.cad_obj is None:
            self.setWindowTitle(f"Add a double filter reactor module")
        else:
            self.setWindowTitle(f"Modify double filter reactor module")

        # Build the basic settings tab
        widget_form_basic = self._buildBasicSettings()

        # Build the advanced settings tab
        widget_form_advanced = self._buildAdvancedSettings()

        # Build the IO tab
        widget_io = self._buildIOTab()

        # ----------------- ASSEMBLING ----------------------------------------

        self.vbox_global = QtWidgets.QVBoxLayout(self)

        self.tabs = QtWidgets.QTabWidget(self)

        self.tabs.addTab(widget_form_basic, "Basic")
        self.tabs.addTab(widget_form_advanced, "Advanced")
        self.tabs.addTab(widget_io, "Inputs/Outputs")

        self.ok_button = QtWidgets.QPushButton("OK", self)
        self.del_button = QtWidgets.QPushButton("Delete", self)
        split_butttons = QtWidgets.QHBoxLayout()
        split_butttons.addWidget(self.ok_button)
        split_butttons.addWidget(self.del_button)

        # if new object only display ok button
        if self.cad_obj is None:
            self.del_button.setVisible(False)

        # self.vbox_global.addLayout(fbox)
        self.vbox_global.addWidget(self.tabs)
        self.vbox_global.addLayout(split_butttons)

        self.setLayout(self.vbox_global)
        self.show()

    def _buildBasicSettings(self) -> QtWidgets.QWidget:

        """Build the widget for the basic settings"""

        widget_form_basic = QtWidgets.QWidget(self)
        fbox = QtWidgets.QFormLayout()
        widget_form_basic.setLayout(fbox)

        # Spin box to specify reactor's volume
        # Set default and limit values
        self.spin_vol_top = QtWidgets.QDoubleSpinBox(self)
        self.spin_vol_top.setValue(5)
        self.spin_vol_top.setMinimum(0)
        self.spin_vol_top.setMaximum(100)

        # Spin box to specify reactor's volume
        # Set default and limit values
        self.spin_vol_bottom = QtWidgets.QDoubleSpinBox(self)
        self.spin_vol_bottom.setValue(10)
        self.spin_vol_bottom.setMinimum(0)
        self.spin_vol_bottom.setMaximum(100)

        # Combo box to choose top of reactor
        self.combo_top = QtWidgets.QComboBox(self)
        list_top = df_rea.TOPS
        self.combo_top.addItems(list_top)

        self.combo_bottom = QtWidgets.QComboBox(self)
        list_bottoms = df_rea.BOTTOMS
        self.combo_bottom.addItems(list_bottoms)

        # Spin box to enter the height of the floating filter
        # Set default and limit values
        self.spin_h_filter = QtWidgets.QDoubleSpinBox(self)
        self.spin_h_filter.setValue(3)
        self.spin_h_filter.setMinimum(0)
        self.spin_h_filter.setMaximum(200)

        # Spin box to enter the diameter of the floating filter
        # Set default and limit values
        self.spin_d_filter = QtWidgets.QDoubleSpinBox(self)
        self.spin_d_filter.setValue(20)
        self.spin_d_filter.setMinimum(0)
        self.spin_d_filter.setMaximum(200)

        # Add widgets to form layout
        fbox.addRow("Volume top chamber (mL): ", self.spin_vol_top)
        fbox.addRow("Volume bottom chamber (mL): ", self.spin_vol_bottom)
        fbox.addRow("Top type: ", self.combo_top)
        fbox.addRow("Bottom type: ", self.combo_bottom)
        fbox.addRow("Filters thickness: ", self.spin_h_filter)
        fbox.addRow("Filters diameter: ", self.spin_d_filter)

        return widget_form_basic

    def _buildAdvancedSettings(self) -> QtWidgets.QWidget:

        """Build the widget for the advanced settings"""

        widget_form_advanced = QtWidgets.QWidget(self)
        fbox = QtWidgets.QFormLayout()
        widget_form_advanced.setLayout(fbox)

        # Internal pipe diameter
        # Set default and disable
        self.spin_d_can = QtWidgets.QDoubleSpinBox(self)
        self.spin_d_can.setValue(cst.D_CAN)
        self.spin_d_can.setEnabled(False)

        # Combo box to choose how to align the top of the reactor
        self.combo_align_top_strategy = QtWidgets.QComboBox(self)
        self.combo_align_top_strategy.addItems(df_rea.ALIGN_TOP_STRATEGIES)

        # How to expand the module if expand is chosen to align the top
        self.combo_expand_strategy = QtWidgets.QComboBox(self)
        self.combo_expand_strategy.addItems(df_rea.EXPAND_STRATEGIES)

        # Combo box for align filter strategy
        self.combo_align_filter_strategy = QtWidgets.QComboBox(self)
        self.combo_align_filter_strategy.addItems(df_rea.ALIGN_FILTER_STRATEGIES)

        # Add widgets to form layout
        fbox.addRow("Diameter internal pipe: ", self.spin_d_can)
        fbox.addRow("Align top strategy: ", self.combo_align_top_strategy)
        fbox.addRow("Expand strategy: ", self.combo_expand_strategy)
        fbox.addRow("Align filter strategy: ", self.combo_align_filter_strategy)

        return widget_form_advanced

    def _buildIOTab(self) -> QtWidgets.QWidget:

        """Build the IO tab"""

        # I/O widget with layout
        widget_io = QtWidgets.QWidget(self)
        vbox_io = QtWidgets.QVBoxLayout()
        widget_io.setLayout(vbox_io)

        self.table_io = QtWidgets.QTableWidget(0, 3, self)
        self.table_io.setHorizontalHeaderLabels(["Name", "Type", "Connected"])
        self.table_io.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)

        # Stretch header to fill entire space
        self.table_io.horizontalHeader().setSectionResizeMode(
            QtWidgets.QHeaderView.Stretch
        )

        # Disable header resizing
        self.table_io.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Fixed)

        # Selection selects entire rows
        self.table_io.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)

        # Only select one row at a time
        self.table_io.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)

        # Button to add io (opens a dialog box)
        self.button_add_io = QtWidgets.QPushButton("Add I/O", self)

        # Button to remove selected io
        self.button_delete_io = QtWidgets.QPushButton("Delete I/O", self)

        # Add the widgets to the layout
        vbox_io.addWidget(self.table_io)
        vbox_io.addWidget(self.button_add_io)
        vbox_io.addWidget(self.button_delete_io)

        return widget_io


if __name__ == "__main__":
    print(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
    app = QtWidgets.QApplication(sys.argv)
    obj = ChemModule()
    sys.exit(app.exec_())



#!/usr/bin/python
# coding: utf-8

import os
import pytest
import numpy as np
import itertools
import subprocess as sp
import trimesh
import solid.utils as su

from ccad.log import MyLog
from ccad.reactor import ReactorCAD as rea
from ccad.siphon import SiphonCAD
import ccad.constants as cst
from ccad.exceptions import *


l = MyLog("output_tests_siphon.log", mode="w")

l.debug("---------------------- START NEW RUN OF TESTS ----------------------")
l.debug("\n")


def logAssert(test, msg):

    """
    Function to log the result of an assert
    http://stackoverflow.com/questions/24892396/py-test-logging-messages-and-test-results-assertions-into-a-single-file
    """

    if not test:
        l.error(msg)
        assert test, msg


def test_error_handling():

    """
    Siphon class must crash if instantiated with an input and output w
    different diameters, or if input/output are too close to each other,
    or any invalid parameter
    """

    l.debug("Testing error handling")

    r_a = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)
    r_b = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)

    # Create 2 inputs at the same height
    r_a.addOutputPercentage("test")
    r_b.addInputPercentage("test")

    # Check that exp is raised if I/O vertically too close
    with pytest.raises(ConstraintError):
        SiphonCAD(r_a, r_a.outputs["test"], r_b, r_b.inputs["test"])
        l.error("Siphon: no exception raised when I/O vertically too close")

    r_a = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)
    r_b = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)

    # Create 2 inputs with different diameters
    r_a.addOutputPercentage("test", height_per=20, diameter=2)
    r_b.addInputPercentage("test", diameter=3)

    # Check that exp is raised if I/O have different diameters
    with pytest.raises(IncompatibilityError):
        SiphonCAD(r_a, r_a.outputs["test"], r_b, r_b.inputs["test"])
        l.error("Siphon: I/O with different diameters")


def test_properties():

    """Make sure properties (setters/getters) work properly"""

    # TODO: tests on offsets_type

    l.debug("Testing properties")

    r_a = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)
    r_b = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)

    # Create 2 inputs with different diameters
    r_a.addOutputPercentage("test", height_per=20)
    r_b.addInputPercentage("test")

    c = SiphonCAD(r_a, r_a.outputs["test"], r_b, r_b.inputs["test"])

    # Test values returned for properties
    logAssert(
        np.isclose(c.length, 10.33, 1e-2),
        "Properties: wrong length {} returned".format(c.length),
    )

    logAssert(c.width == 9, "Properties: wrong width {} returned".format(c.width))

    logAssert(
        np.isclose(c.height, 36.64, 1e-2),
        "Properties: wrong height returned: {}".format(c.height),
    )

    c.length = 20

    x = np.array(r_b.coo["x"])

    logAssert(c.length == 20, "Properties: wrong length after modif")

    # After modifying length, 2nd reactor must have moved
    logAssert(
        np.isclose(x, -49.73, 1e-2),
        "Properties: wrong x for 2nd reactor after modif length {}".format(x),
    )


def test_translations():

    """
    Check that the siphon will properly translate the objects it connects
    """

    r_a = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)
    r_b = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)

    r_b.addInputPercentage("test")

    c = SiphonCAD(r_a, r_a.outputs["default"], r_b, r_b.inputs["test"])

    x = np.array(r_b.coo["x"])

    # First test, 2 reactors aligned on x axis
    logAssert(
        np.isclose(x, 40.06, 1e-2), "Translation: wrong x {} for 2nd reactor".format(x)
    )

    r_a = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O, 30)
    r_b = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)

    r_b.addInputPercentage("test")

    c = SiphonCAD(r_a, r_a.outputs["default"], r_b, r_b.inputs["test"])

    x = np.array(r_b.coo["x"])
    y = np.array(r_b.coo["y"])

    # 2 reactors, first one has its output at 30°
    logAssert(
        np.isclose(x, 34.69, 1e-2),
        "Translation: wrong x {} for 2nd reactor, output 1st at 30°".format(x),
    )
    logAssert(
        np.isclose(y, 20.03, 1e-2),
        "Translation: wrong y {} for 2nd reactor, output 1st at 30°".format(y),
    )

    r_a = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O, 30)
    r_b = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O, 0)
    r_c = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O, 0)

    r_b.addInputPercentage("test", angle=-30)
    r_c.addInputPercentage("test")

    c = SiphonCAD(r_a, r_a.outputs["default"], r_b, r_b.inputs["test"])
    c_2 = SiphonCAD(r_b, r_b.outputs["default"], r_c, r_c.inputs["test"])

    x = np.array(r_c.coo["x"])
    y = np.array(r_c.coo["y"])
    angle = np.array(r_c.coo["angle"])

    # 3 reactors, first one has its output at 30°, 2nd one has its input at
    # -30° (30 ° from its output). Results in collision between 1st and 3rd
    # reactor
    logAssert(
        np.isclose(x, 14.66, 1e-2),
        "Translation: wrong x {}for 3rd reactor, ultimate test".format(x),
    )
    logAssert(
        np.isclose(y, -14.66, 1e-2),
        "Translation: wrong y {} for 3rd reactor, ultimate test".format(y),
    )
    logAssert(
        np.isclose(angle, 240, 1e-2),
        "Translation: wrong angle {}, ultimate test".format(angle),
    )


def test_valid_stl():

    """
    Render the STL file for 2 reactors, one siphon
    Then test the validity of the STL file:
    - Check if watertight (True if there is non non-manifold surfaces)
    - Check valid volume
    """

    l.debug("Testing STLs validity")

    r_a = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)
    r_b = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)

    r_b.addInputPercentage("test")

    c = SiphonCAD(r_a, r_a.outputs["default"], r_b, r_b.inputs["test"])

    su.scad_render_to_file(r_a.cad + r_b.cad + c.cad, "test.scad")

    # Render the reactor into a stl file
    cmd = sp.Popen(
        ["openscad", "-o", "test.stl", "test.scad"], stdout=sp.PIPE, stderr=sp.PIPE
    )

    # Wait until rendering is finished
    out, err = cmd.communicate()

    # Check if rendering succeeded, crash the test otherwise
    if cmd.returncode != 0:
        l.error("Failed to render STL for 2 reactors, 1 siphon")
        pytest.fail("Failed to render STL for 2 reactors, 1 siphon")

    # Generate the mesh from STL file
    mesh = trimesh.load_mesh("test.stl")

    logAssert(mesh.is_watertight is True, "2 reactors, 1 siphon, not watertight")

    logAssert(mesh.volume > 0, "2 reactors, 1 siphon, negative volume")

    # Remove files to not polute dir
    os.remove("test.scad")
    os.remove("test.stl")


if __name__ == "__main__":
    pass



#!/usr/bin/python
# coding: utf-8

import pytest
import numpy as np
import itertools
import subprocess as sp
import trimesh
import os

from ccad.log import MyLog
from ccad.reactor import ReactorCAD as rea
from ccad.filter_reactor import FilterReactorCAD as f_rea
from ccad.top_in import TopInlet
from ccad.exceptions import *


l = MyLog("output_tests_filter_reactor.log", mode="w")

l.debug("---------------------- START NEW RUN OF TESTS ----------------------")
l.debug("\n")


def logAssert(test, msg):

    """
    Function to log the result of an assert
    http://stackoverflow.com/questions/24892396/py-test-logging-messages-and-test-results-assertions-into-a-single-file
    """

    if not test:
        l.error(msg)
        assert test, msg


@pytest.fixture()
def all_reactors() -> list:

    """
    Fixture:
    return all types of reactors, based on permutations of tops and bottoms
    """

    list_reactors = []

    tops = rea.TOPS
    bottoms = rea.BOTTOMS

    # Generate a list of all possible permutations
    for couple in itertools.product(tops, bottoms):
        list_reactors.append((couple[0], couple[1]))

    l.debug("List of reactors generated:\n")
    l.debug(list_reactors)

    return list_reactors


def test_error_handling():

    """Reactor class must crash if provided with invalid parameters"""

    l.debug("Testing error handling")

    # Raise exception if simple round bottom selected
    with pytest.raises(NotImplementedError):
        f_rea(20, rea.T_CLOSED_ROUND, rea.B_ROUND, d_filter=20, h_filter=3)
        l.error("f_rea class: no exception raised when round bottom")

    # Raise excpetion when constrained radius is too small for filter
    with pytest.raises(ConstraintError):
        f_rea(10, rea.T_2_QUA_INCH, rea.B_FLAT_IN_O, d_filter=30, h_filter=3)
        l.error("f_rea class: no exp when constrained r too small")


def test_h_r_filter_too_small(all_reactors: list):

    """
    For a 20 mL reactor, check that h and R are right, depending on the
    shape of the reactor. Filter is too small for these tests
    """

    l.debug("Testing h and r")

    for couple in all_reactors:

        l.debug(
            "Testing h and r for couple Top: {}; Bottom: {}".format(
                couple[0], couple[1]
            )
        )

        # Round bottom is passed, not possible
        if couple[1] == rea.B_ROUND:
            continue

        r = f_rea(20, couple[0], couple[1], d_filter=20, h_filter=3)

        # Closed top, round bottom (without or with inlet)
        if couple[0] in rea.CLOSED_TOPS and couple[1] in rea.ROUND_BOTS:

            logAssert(
                np.isclose(r._h, 36.41, 1e-2),
                "h {} is wrong for couple {}".format(r._h, couple),
            )
            logAssert(
                np.isclose(r._r, 11.25, 1e-2),
                "R {} is wrong for couple {}".format(r._r, couple),
            )

        elif couple[0] == rea.T_OPEN and couple[1] in rea.ROUND_BOTS:

            logAssert(
                np.isclose(r._h, 38.36, 1e-2),
                "h {} is wrong for couple {}".format(r._h, couple),
            )
            logAssert(
                np.isclose(r._r, 11.85, 1e-2),
                "R {} is wrong for couple {}".format(r._r, couple),
            )

        # Closed tops, flat bottoms (with or without inlet)
        elif couple[0] in rea.CLOSED_TOPS and couple[1] in rea.FLAT_BOTS:

            logAssert(
                np.isclose(r._h, 37.97, 1e-2),
                "h {} is wrong for couple {}".format(r._h, couple),
            )
            logAssert(
                np.isclose(r._r, 11.73, 1e-2),
                "R {} is wrong for couple {}".format(r._r, couple),
            )

        # Open top, flat bottom (without or with inlet)
        elif couple[0] in rea.T_OPEN and couple[1] in rea.FLAT_BOTS:

            logAssert(
                np.isclose(r._h, 40.41, 1e-2),
                "h {} is wrong for couple {}".format(r._h, couple),
            )
            logAssert(
                np.isclose(r._r, 12.49, 1e-2),
                "R {} is wrong for couple {}".format(r._r, couple),
            )

        # 1/2 inch threaded top, flat bottom
        elif couple[0] == rea.T_2_QUA_INCH and couple[1] in rea.FLAT_BOTS:

            logAssert(
                np.isclose(r._h, 63.02, 1e-2),
                "h {} is wrong for couple {}".format(r._h, couple),
            )
            logAssert(r._r == 10, "R {} is wrong for couple {}".format(r._r, couple))

        # 1/2 inch threaded top, round bottom
        elif couple[0] == rea.T_2_QUA_INCH and couple[1] in rea.ROUND_BOTS:

            logAssert(
                np.isclose(r._h, 58.80, 1e-2),
                "h {} is wrong for couple {}".format(r._h, couple),
            )
            logAssert(r._r == 10, "R {} is wrong for couple {}".format(r._r, couple))

        # 3/4 inch threaded top, flat bottom
        elif couple[0] == rea.T_3_QUA_INCH and couple[1] in rea.FLAT_BOTS:

            logAssert(
                np.isclose(r._h, 43.77, 1e-2),
                "h {} is wrong for couple {}".format(r._h, couple),
            )
            logAssert(r._r == 12, "R {} is wrong for couple {}".format(r._r, couple))

        # 1/2 inch threaded top, round bottom
        elif couple[0] == rea.T_3_QUA_INCH and couple[1] in rea.ROUND_BOTS:

            logAssert(
                np.isclose(r._h, 37.11, 1e-2),
                "h {} is wrong for couple {}".format(r._h, couple),
            )
            logAssert(r._r == 12, "R {} is wrong for couple {}".format(r._r, couple))

        # 1 inch threaded top, flat bottom
        elif couple[0] == rea.T_4_QUA_INCH and couple[1] in rea.FLAT_BOTS:

            logAssert(
                np.isclose(r._h, 24.62, 1e-2),
                "h {} is wrong for couple {}".format(r._h, couple),
            )
            logAssert(r._r == 16, "R {} is wrong for couple {}".format(r._r, couple))

        # 1/2 inch threaded top, round bottom
        elif couple[0] == rea.T_4_QUA_INCH and couple[1] in rea.ROUND_BOTS:

            logAssert(
                np.isclose(r._h, 14.54, 1e-2),
                "h {} is wrong for couple {}".format(r._h, couple),
            )
            logAssert(r._r == 16, "R {} is wrong for couple {}".format(r._r, couple))

        else:
            l.error("output: unrecognized couple: {}; {}".format(couple[0], couple[1]))
            pytest.fail(
                "Output test failed for unrecognized couple: {}; {}".format(
                    couple[0], couple[1]
                )
            )


def test_h_r_filter_not_too_small(all_reactors: list):

    """
    For a 10 mL reactor, check that h and R are right, depending on the
    shape of the reactor. Filter is NOT too small for these tests
    """

    l.debug("Testing h and r")

    for couple in all_reactors:

        # Pass simple round bottom, not possible at all
        if couple[1] == rea.B_ROUND:
            continue

        # Pass threaded tops, radius is constrained and 20mm filter is
        # too small
        if couple[0] in rea.THREADED_TOPS:
            continue

        l.debug(
            "Testing h and r for couple Top: {}; Bottom: {}".format(
                couple[0], couple[1]
            )
        )

        # Test exception is raised if volume is too small for round bottoms
        if couple[1] in rea.ROUND_BOTS:
            with pytest.raises(ImpossibleAction):
                r = f_rea(10, couple[0], couple[1], d_filter=20, h_filter=3)
                l.error(
                    "f_rea: no exp w/ round bot & filter not too small {}".format(
                        couple
                    )
                )
            continue

        r = f_rea(10, couple[0], couple[1], d_filter=20, h_filter=3)

        # Closed top, round bottom (without or with inlet)
        # This block will never be executed, round bottoms when filter is
        # not too small are not permitted
        if couple[0] in rea.CLOSED_TOPS and couple[1] in rea.ROUND_BOTS:
            pass

        # Open top, round bottoms
        elif couple[0] == rea.T_OPEN and couple[1] in rea.ROUND_BOTS:

            logAssert(
                np.isclose(r._h, 38.36, 1e-2),
                "h {} is wrong for couple {}".format(r._h, couple),
            )
            logAssert(
                np.isclose(r._r, 11.85, 1e-2),
                "R {} is wrong for couple {}".format(r._r, couple),
            )

        # Closed tops, flat bottoms (with or without inlet)
        elif couple[0] in rea.CLOSED_TOPS and couple[1] in rea.FLAT_BOTS:

            logAssert(
                np.isclose(r._h, 42.23, 1e-2),
                "h {} is wrong for couple {}".format(r._h, couple),
            )
            logAssert(
                np.isclose(r._r, 8.43, 1e-2),
                "R {} is wrong for couple {}".format(r._r, couple),
            )

        # Open top, flat bottom (without or with inlet)
        elif couple[0] in rea.T_OPEN and couple[1] in rea.FLAT_BOTS:

            logAssert(
                np.isclose(r._h, 47.84, 1e-2),
                "h {} is wrong for couple {}".format(r._h, couple),
            )
            logAssert(
                np.isclose(r._r, 8.43, 1e-2),
                "R {} is wrong for couple {}".format(r._r, couple),
            )

        else:
            l.error("output: unrecognized couple: {}; {}".format(couple[0], couple[1]))
            pytest.fail(
                "Output test failed for unrecognized couple: {}; {}".format(
                    couple[0], couple[1]
                )
            )


def test_infos():

    """Make sure the dict 'infos' returns proper values"""

    l.debug("Testing filter reactor's infos")

    # Testing with filter too small
    r = f_rea(20, rea.T_OPEN, rea.B_ROUND_IN_O, 20, 3)

    logAssert(r.infos["z_top_filter"] == 12, "infos: wrong z_top_filter")
    logAssert(r.infos["z_bottom_filter"] == 9, "infos: wrong z_top_filter")
    logAssert(r.infos["h_filter"] == 3, "infos: wrong h_filter")
    logAssert(r.infos["d_filter"] == 20, "infos: wrong d_filter")
    logAssert(r.infos["filter_too_small"] is True, "infos: wrong filter_too_small")

    # Testing with filter NOT too small
    r = f_rea(10, rea.T_OPEN, rea.B_FLAT_IN_O, 20, 3)

    logAssert(r.infos["z_top_filter"] == 12, "infos 2): wrong z_top_filter")
    logAssert(r.infos["z_bottom_filter"] == 9, "infos 2): wrong z_top_filter")
    logAssert(r.infos["h_filter"] == 3, "infos 2): wrong h_filter")
    logAssert(r.infos["d_filter"] == 20, "infos 2): wrong d_filter")
    logAssert(r.infos["filter_too_small"] is False, "infos 2): wrong filter_too_small")


def test_valid_stl(all_reactors: list):

    """
    Generate all possible shapes of reactor, and render the STL files.
    Then test the validity of each STL file:
    - Check if watertight (True if there is non non-manifold surfaces)
    - Check if valid volume
    - Check if valid winding
    """

    l.debug("Testing STLs validity")

    for couple in all_reactors:

        l.debug(
            "Building and rendering couple Top: {}; Bottom: {}".format(
                couple[0], couple[1]
            )
        )

        # Round bottom is passed, not possible
        if couple[1] == rea.B_ROUND:
            continue

        # Build the reactor
        r = f_rea(20, couple[0], couple[1], 20, 3)

        # Add inputs to check for non-manifolds
        r.addInputPercentage("i1", 0)
        r.addInputPercentage("i2", 100)

        # Forge the name of the file
        name = "{} {}".format(couple[0], couple[1])
        name = name.replace(" ", "_")
        name = name.replace("/", "_")
        r.renderToFile(name + ".scad")

        # Render the reactor into a stl file
        cmd = sp.Popen(
            ["openscad", "-o", name + ".stl", name + ".scad"],
            stdout=sp.PIPE,
            stderr=sp.PIPE,
        )

        # Wait until rendering is finished
        out, err = cmd.communicate()

        # Check if rendering succeeded, crash the test otherwise
        if cmd.returncode != 0:
            l.error(
                "Failed to render STL for couple Top {}; Bottom: {}".format(
                    couple[0], couple[1]
                )
            )
            pytest.fail(
                "Failed to render STL for couple Top: {}; Bottom: {}".format(
                    couple[0], couple[1]
                )
            )

        # Generate the mesh from STL file
        mesh = trimesh.load_mesh(name + ".stl")

        logAssert(
            mesh.is_watertight,
            "Couple Top: {}; Bottom: {} has non-manifold surfaces".format(
                couple[0], couple[1]
            ),
        )

        logAssert(
            mesh.is_volume,
            "Couple Top: {}; Bottom: {} has a negative volume".format(
                couple[0], couple[1]
            ),
        )

        logAssert(
            mesh.is_winding_consistent,
            "Couple Top: {}; Bottom: {} has unconsistent winding".format(
                couple[0], couple[1]
            ),
        )

        # Remove files to not polute dir
        os.remove(name + ".scad")
        os.remove(name + ".stl")


if __name__ == "__main__":
    pass



#!/usr/bin/python
# coding: utf-8

# https://stackoverflow.com/questions/12576454/python-and-matplotlib-and-annotations-with-mouse-hover
# https://stackoverflow.com/questions/7908636/possible-to-make-labels-appear-when-hovering-over-a-point-in-matplotlib


import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from PyQt5 import QtCore, QtWidgets
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigCanvas
from matplotlib.figure import Figure
from typing import Dict, Tuple, List

# matplotlib.use("Qt5Agg")

from ccad.object import ObjectCAD
from ccad import ReactorCAD
import chemscad.app_constants as app_cst


class Dashboard(FigCanvas):

    """
    A canvas. Will be used as a dashboard to display the assembly. Has mouse hovering
    and mouse clicking feature.  Snake case is used in this module for compatibility
    purposes with matplotlib function names. Methods called by GUI will use Camel case
    """

    def __init__(
        self, parent=None, width: int = 5, height: int = 4, dpi: int = 100
    ) -> None:

        fig = Figure(figsize=(width, height), dpi=dpi)
        super(Dashboard, self).__init__(fig)
        gs = mpl.gridspec.GridSpec(2, 1, height_ratios=[1, 3])
        self.axes = fig.add_subplot(gs[1])
        self.axes_info = fig.add_subplot(gs[0])

        # Remove axes
        self.axes.axis("off")
        self.axes_info.axis("off")

        # Initialise info text
        self.info_text = self.axes_info.text(0, 0, "")

        # # Display grid
        self.axes.grid(True)

        # self.setParent(parent)
        self.parent = parent

        FigCanvas.setSizePolicy(
            self, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        )
        FigCanvas.updateGeometry(self)

        # Connect the event to internal callbacks: mouse hovering and
        # mouse clicking
        self.mpl_connect("motion_notify_event", self._on_move)
        fig.canvas.mpl_connect("pick_event", self._on_pick)

        # Dict to associate a patch (circle, rectangle, etc) to a cad object
        self.dict_buttons: Dict[plt.Circle, Tuple[ReactorCAD, str]] = dict()

        self.x_min = 0
        self.y_min = 0

    def _draw_reactors(self):

        """Methods to draw the reactor-like objects (not the connectors)"""

        # Store unconnected reactors
        list_unconnected: List[ReactorCAD] = list()

        for index, reactor in enumerate(self.parent.ass.list_reactors):

            # If the reactor is not connected, don't draw it now.
            # Store it and deal with it later
            if not reactor.isConnected():
                list_unconnected.append(reactor)
                continue

            # Get the coordinates of the reactor and its radius
            x, y, r = reactor.coo["x"], reactor.coo["y"], reactor.infos["r_ex"]

            # Get proper colour for module
            colour = self._get_module_colour(reactor.module_type)

            # Create a circle at the coo of the reactor, with r of the reactor.
            # Picker is True to activate mouse click events
            circle = plt.Circle((x, y), radius=r, picker=True, fc=colour, ec="k")

            # Rotate the circle, for aesthetics purposes
            # (the CAD process expands the assembly towards -y, rotate to
            # expand towards +x)
            t = mpl.transforms.Affine2D().rotate_deg(-90) + self.axes.transData
            circle.set_transform(t)

            # Calculate (x, y) for input of input object
            # Calculate new x,y coordinates by using a transformation matrix
            # rotation of -90°
            a = np.deg2rad(-90)
            rotation = np.array([[np.cos(a), -np.sin(a)], [np.sin(a), np.cos(a)]])
            coo = np.array([x, y])

            x, y = np.dot(rotation, coo)

            # Find short name for reactor. Use assembly
            for name, module in self.parent.ass.reactors_names.items():
                if reactor is module:
                    module_name = name

            self.axes.annotate(
                module_name, xy=(x, y), ha="center", va="center", fontsize=20
            )

            # Associate the circle to the CAD object, and to its name
            self.dict_buttons[circle] = (reactor, reactor.module_type)

            # Add the circle to the axes
            self.axes.add_patch(circle)

        self._update_figure()

        # Now draw unconnected reactors separately
        self._draw_unconnected_reactor(list_unconnected)

    def _draw_unconnected_reactor(self, list_unconnected: List[ReactorCAD]):

        """
        Method to draw the unconnected reactors. Called from _draw_reactors
        """

        # Deal with the unconnected reactors
        for reactor in list_unconnected:

            # index = self.parent.ass.list_reactors.index(reactor)

            # Get x axis max limit
            _, x_max = self.axes.get_xlim()

            # Assign coordinates to unconnected reactor (don't simply use
            # (0, 0))
            x, y, r = x_max, 0, reactor.infos["r_ex"]

            # Add some space in x before drawing reactor
            x = x + 2 * r

            # Get proper colour for module
            colour = self._get_module_colour(reactor.module_type)

            # Create a circle at the coo of the reactor, with r of the reactor.
            # Picker is True to activate mouse click events
            # circle = plt.Circle((x, y), radius=r, picker=True)
            circle = plt.Circle((x, y), radius=r, picker=True, fc=colour, ec="k")

            # Find short name for reactor. Use assembly
            for name, module in self.parent.ass.reactors_names.items():
                if reactor is module:
                    module_name = name

            self.axes.annotate(
                module_name, xy=(x, y), ha="center", va="center", fontsize=20
            )

            # Associate the circle to the CAD object, and to its name
            self.dict_buttons[circle] = (reactor, reactor.module_type)

            # Add the circle to the axes
            self.axes.add_patch(circle)

            # Update the graph in the loop, to always get x_max right after
            # adding an unconnected reactor
            self._update_figure()

    def _draw_connectors(self):

        """Methods to draw the connectors objects"""

        for con in self.parent.ass.list_connectors:

            x, y, angle = con.coo["x"], con.coo["y"], con.coo["angle"]
            length, width = con.length, con.width

            # Translate rectangles properly (origin of rectangle is lower
            # left corner)
            x = x + (width / 2) * np.sin(np.deg2rad(abs(angle)))
            y = y - (width / 2) * np.cos(np.deg2rad(abs(angle)))

            # Get proper colour for module
            colour = self._get_module_colour(con.module_type)

            rect = plt.Rectangle(
                (x, y), length, width, fc=colour, angle=angle, picker=True, ec="k"
            )

            # Rotate the rectangle, for aesthetics purposes
            # (the CAD process expands the assembly towards -y, rotate to
            # expand towards +x)
            t = mpl.transforms.Affine2D().rotate_deg(-90) + self.axes.transData
            rect.set_transform(t)

            # Add the rectangle to the axes
            self.axes.add_patch(rect)

            # Associate the rect to the CAD object, and to its name
            self.dict_buttons[rect] = (con, con.module_type)

        self._update_figure()

    def _get_module_colour(self, module_type: str) -> str:

        """
        Return the colour for `module_type`. Uses a dict defined in the
        app_constants. If the module type is unknown, use white. Return a
        hexadecimal colour
        """

        dict_colours = app_cst.colours_modules

        # Get colour for module. If unknown module, use white
        hex_colour = dict_colours.get(module_type, "#FFFFFF")

        return hex_colour

    def _clear_figure(self):

        """Clear the graph"""

        self.axes.clear()
        self._update_figure()

    def _update_figure(self):

        """Update the graph. Necessary, to call after each plot"""

        # Reset the axes aspect
        self.axes.set_aspect("equal", adjustable="box")
        self.axes.autoscale_view()
        self.axes.axis("off")

        self.draw()

    def _clear_info_text(self):

        """Clear info text subplot"""

        self.info_text.set_text("")

    def _on_move(self, event):

        """
        Method to handle mouse hovering. Subclassed from matplotlib Canvas
        """

        self._clear_info_text()
        # For all the pacthes in the canvas
        for patch, couple in self.dict_buttons.items():

            # TODO: different colours depending on module_type
            module_type = couple[1]

            # If the pacth contains the mouse, change colour of patch to
            # show selection. Also change colour of all the other patches,
            # to show unselection
            if patch.contains(event)[0]:
                self._clear_info_text()
                self.info_text.set_text(self._get_info_text(couple[0]))
                patch.set_facecolor("#FF6962")
            else:
                colour = self._get_module_colour(module_type)
                patch.set_facecolor(colour)
        self._update_figure()

    def _on_pick(self, event):

        """
        Method to handle click event on the pacthes. Subclassed from
        matplotlib Canvas
        """

        # From the patch clicked, get the CAD object and its type
        cad_obj, module_type = self.dict_buttons[event.artist]

        # Open the dialog box corresponding to the type of the CAD object
        self.parent.openModuleDialog(cad_obj)

    def refresh(self):

        """
        Refresh the canvas. Called by parent (GUI) when assembly is modified
        """

        # Reset the dict of objects, will be rebuilt
        del self.dict_buttons
        self.dict_buttons = dict()

        self.x_min = 0
        self.y_min = 0

        # Clear the canvas
        self._clear_figure()

        # Draw the reactors
        self._draw_reactors()

        # Draw the connectors
        self._draw_connectors()

    def _get_info_text(self, module: ObjectCAD):

        """Return str presenting info of given module"""

        infos = module.infos
        info_text = ""

        for key, value in infos.items():

            # v_str is the vlaue properly formatted (2 decimals if float)
            if isinstance(value, float):
                v_str = f"{value:.2f}"
            else:
                v_str = str(value)

            info_text += f"{key}: {v_str}\n"

        return info_text



#!/usr/bin/python
# coding: utf-8

from PyQt5 import QtCore


class LittleThread(QtCore.QThread):

    """Create a personal thread API, to spawn a thread with
    any function and its parameters"""

    # https://joplaete.wordpress.com/2010/07/21/threading-with-pyqt4/
    # http://stackoverflow.com/questions/34496313/display-busy-progressbar-for-long-process-no-thread/34496585#34496585

    def __init__(self, function, *args, **kwargs):

        QtCore.QThread.__init__(self)

        self.function = function
        self.args = args
        self.kwargs = kwargs

    def run(self):

        self.function(*self.args, **self.kwargs)
        return



#!/usr/bin/python
# coding: utf-8

import pytest
import numpy as np
import itertools
import subprocess as sp
import trimesh
import os

from ccad.log import MyLog
from ccad.reactor import ReactorCAD as rea
from ccad.floating_filter_reactor import FloatingFilterReactorCAD as ffr
from ccad.exceptions import ConstraintError


l = MyLog("output_tests_floating_filter_reactor.log", mode="w")

l.debug("---------------------- START NEW RUN OF TESTS ----------------------")
l.debug("\n")


def logAssert(test, msg):

    """
    Function to log the result of an assert
    http://stackoverflow.com/questions/24892396/py-test-logging-messages-and-test-results-assertions-into-a-single-file
    """

    if not test:
        l.error(msg)
        assert test, msg


@pytest.fixture()
def all_reactors() -> list:

    """
    Fixture:
    return all types of reactors, based on permutations of tops and bottoms
    """

    list_reactors = []

    tops = rea.TOPS
    bottoms = rea.BOTTOMS

    # Generate a list of all possible permutations
    for couple in itertools.product(tops, bottoms):
        list_reactors.append((couple[0], couple[1]))

    l.debug("List of reactors generated:\n")
    l.debug(list_reactors)

    return list_reactors


def test_volume_constraints():

    l.debug("Testing top and bottom volume constraints")

    # Negative top volume
    with pytest.raises(ConstraintError):
        ffr(10, -1, ffr.T_CLOSED_ROUND, ffr.B_FLAT_EX_O, d_filter=10, h_filter=10)
        l.error("No exception raised for negative top volume")

    # Negative bottom volume
    with pytest.raises(ConstraintError):
        ffr(-1, 20, ffr.T_CLOSED_ROUND, ffr.B_FLAT_EX_O, d_filter=10, h_filter=10)
        l.error("No exception raised for negative bottom volume")


def test_coherent_filter_value_properties():

    l.debug("Testing z filter coherence")

    t = ffr(10, 20, ffr.T_CLOSED_ROUND, ffr.B_FLAT_EX_O, d_filter=10, h_filter=10)

    test_z_top_filter = t.z_bottom_filter + t.h_filter
    test_z_bottom_filter = t.z_top_filter - t.h_filter

    logAssert(
        t.z_top_filter == test_z_top_filter,
        "Failed: Test: {} Actual: {}".format(test_z_top_filter, t.z_top_filter),
    )

    logAssert(
        t.z_bottom_filter == test_z_bottom_filter,
        "Failed: Test: {} Actual: {}".format(test_z_bottom_filter, t.z_bottom_filter),
    )


def test_volumetric_units():

    l.debug("Testing volumes are in mL and not mm3")

    t = ffr(10, 20, ffr.T_CLOSED_ROUND, ffr.B_FLAT_EX_O, d_filter=10, h_filter=10)

    logAssert(
        t.volume_bottom == 10, "Failed: Test: {} Actual: {}".format(10, t.volume_bottom)
    )

    logAssert(
        t.volume_bottom == 10, "Failed: Test: {} Actual: {}".format(20, t.volume_top)
    )


def test_d_h_filter_values():

    l.debug("Testing D and H filter values match when created")

    t = ffr(10, 20, ffr.T_CLOSED_ROUND, ffr.B_FLAT_EX_O, d_filter=10, h_filter=10)

    logAssert(t.d_filter == 10, "Failed: Test: {} Actual: {}".format(10, t.d_filter))

    logAssert(t.h_filter == 10, "Failed: Test: {} Actual: {}".format(10, t.h_filter))

    t2 = ffr(10, 20, ffr.T_CLOSED_ROUND, ffr.B_FLAT_EX_O, d_filter=15, h_filter=20)

    logAssert(t2.d_filter == 15, "Failed: Test: {} Actual: {}".format(15, t2.d_filter))

    logAssert(t2.h_filter == 20, "Failed: Test: {} Actual: {}".format(20, t2.h_filter))


def test_radius_raises_exception():

    l.debug("Testing setting r property raises error")

    # Check radius setting
    with pytest.raises(NotImplementedError):
        t = ffr(10, 20, ffr.T_CLOSED_ROUND, ffr.B_FLAT_EX_O, d_filter=10, h_filter=10)
        t.r = 50
        l.error("Expected exception not thrown for setting radius")


def test_percentage_raises_exception():

    l.debug("Testing output and input percentage raises error")

    # Adding input percentage
    with pytest.raises(AttributeError):
        t = ffr(10, 20, ffr.T_CLOSED_ROUND, ffr.B_FLAT_EX_O, d_filter=10, h_filter=10)
        t.addInputPercentage("TEST", 75)

    # Addint output percentage
    with pytest.raises(AttributeError):
        t = ffr(10, 20, ffr.T_CLOSED_ROUND, ffr.B_FLAT_EX_O, d_filter=10, h_filter=10)
        t.addOutputPercentage("TEST", 50)


def test_volume_accuracy():

    l.debug("Testing volume = volume_top + volume_bottom")

    t = ffr(10, 20, ffr.T_CLOSED_ROUND, ffr.B_FLAT_EX_O, d_filter=10, h_filter=10)

    total_volume = t.volume_top + t.volume_bottom

    logAssert(
        t.volume == total_volume,
        "Calculated volume does not match Test: {} Actual: {}".format(
            t.volume, total_volume
        ),
    )


def test_io_percentage_bottom():

    l.debug("Testing accuracy of adding IO% @ bottom")

    t = ffr(6, 4, ffr.T_CLOSED_ROUND, ffr.B_FLAT_EX_O, d_filter=20, h_filter=3)

    # Input Bottom 0%
    l.debug("Testing input percentage bottom 0%")
    t.addInputPercentageBottom("I_BOTTOM_0", 0)
    logAssert("I_BOTTOM_0" in t.inputs.keys(), "Input 0%% failed to create")

    io1 = t.inputs["I_BOTTOM_0"]
    t_dict = {
        "height": 13.5,
        "diameter": 3,
        "chamber": "bottom",
        "angle": 180.0,
        "external": False,
    }

    logAssert(io1.infos == t_dict, f"Dictionaries do not match, got: {io1.infos}")

    # Input Bottom 100%
    l.debug("Testing input percentage bottom 100%")

    t.addInputPercentageBottom("I_BOTTOM_100", 100)
    logAssert("I_BOTTOM_100" in t.inputs.keys(), "Input 100%% failed to create")

    io2 = t.inputs["I_BOTTOM_100"]
    t_dict = {
        "external": False,
        "angle": 180.0,
        "diameter": 3,
        "chamber": "bottom",
        "height": 37.406769517776766,
    }

    logAssert(io2.infos == t_dict, f"Dictionaries do not match, got: {io2.infos}")

    # Output Bottom 0%
    l.debug("Testing Output percentage bottom 0%")
    t.addOutputPercentageBottom("O_BOTTOM_0", 0)
    logAssert("O_BOTTOM_0" in t.outputs.keys(), "Output 0%% failed to create")

    io3 = t.outputs["O_BOTTOM_0"]
    t_dict = {
        "chamber": "bottom",
        "external": False,
        "height": 13.5,
        "diameter": 3,
        "angle": 180.0,
    }

    logAssert(io3.infos == t_dict, f"Dictionaries do not match, got: {io3.infos}")

    # Output Bottom 100%
    l.debug("Testing Output percentage bottom 100%")
    t.addOutputPercentageBottom("O_BOTTOM_100", 100)
    logAssert("O_BOTTOM_100" in t.outputs.keys(), "Output 100%% failed to create")

    io4 = t.outputs["O_BOTTOM_100"]
    t_dict = {
        "diameter": 3,
        "external": False,
        "angle": 180.0,
        "chamber": "bottom",
        "height": 37.406769517776766,
    }

    logAssert(io4.infos == t_dict, f"Dictionaries do not match, got: {io3.infos}")


def test_io_percentage_top():

    l.debug("Testing accuracy of adding IO% @ top")
    t = ffr(6, 4, ffr.T_CLOSED_ROUND, ffr.B_FLAT_EX_O, d_filter=20, h_filter=3)

    # Input Top 0%
    l.debug("Testing input percentage top 0%")
    t.addInputPercentageTop("I_TOP_0", 0)
    logAssert("I_TOP_0" in t.inputs.keys(), "Input 0%% failed to create")

    io1 = t.inputs["I_TOP_0"]
    t_dict = {
        "diameter": 3,
        "external": False,
        "angle": 180.0,
        "height": 43.409769517776766,
        "chamber": "top",
    }

    logAssert(io1.infos == t_dict, f"Dictionaries do not match, got: {io1.infos}")

    # Input Top 100%
    l.debug("Testing input percentage top 100%")
    t.addInputPercentageTop("I_TOP_100", 100)
    logAssert("I_TOP_100" in t.inputs.keys(), "Input 100%% failed to create")

    io2 = t.inputs["I_TOP_100"]
    t_dict = {
        "height": 52.72794919629462,
        "chamber": "top",
        "diameter": 3,
        "angle": 180.0,
        "external": False,
    }

    logAssert(io2.infos == t_dict, f"Dictionaries do not match, got: {io2.infos}")

    # Output Top 0%
    l.debug("Testing output percentage top 0%")
    t.addOutputPercentageTop("O_TOP_0", 0)
    logAssert("O_TOP_0" in t.outputs.keys(), "Input 0%% failed to create")

    io3 = t.outputs["O_TOP_0"]
    t_dict = {
        "height": 43.409769517776766,
        "diameter": 3,
        "external": False,
        "chamber": "top",
        "angle": 180.0,
    }

    logAssert(io3.infos == t_dict, f"Dictionaries do not match, got: {io3.infos}")

    # Output Top 100%
    l.debug("Testing output percentage top 100%")
    t.addOutputPercentageTop("O_TOP_100", 100)
    logAssert("O_TOP_100" in t.outputs.keys(), "Input 100%% failed to create")

    io4 = t.outputs["O_TOP_100"]
    t_dict = {
        "external": False,
        "height": 52.72794919629462,
        "chamber": "top",
        "diameter": 3,
        "angle": 180.0,
    }

    logAssert(io4.infos == t_dict, f"Dictionaries do not match, got: {io4.infos}")


def test_h_r(all_reactors: list):

    """
    For a 20 mL reactor, check that h and R are right, depending on the
    shape of the reactor
    """

    l.debug("Testing h and r")

    for couple in all_reactors:

        # Pass threaded tops, radius is constrained by the filter
        if couple[0] in rea.THREADED_TOPS:
            continue

        l.debug(
            "Testing h and r for couple Top: {}; Bottom: {}".format(
                couple[0], couple[1]
            )
        )

        r = ffr(6, 4, couple[0], couple[1], d_filter=20, h_filter=3)

        # Open top, round bottoms OR Closed tops, flat bottoms. Same h and r
        if (
            couple[0] == ffr.T_OPEN
            and couple[1] in ffr.ROUND_BOTS
            or couple[0] in ffr.CLOSED_TOPS
            and couple[1] in ffr.FLAT_BOTS
        ):

            logAssert(
                np.isclose(r._h, 42.23, 1e-2),
                "h {} is wrong for couple {}".format(r._h, couple),
            )
            logAssert(
                np.isclose(r._r, 8.43, 1e-2),
                "R {} is wrong for couple {}".format(r._r, couple),
            )

        # Open top, flat bottoms
        elif couple[0] == ffr.T_OPEN and couple[1] in ffr.FLAT_BOTS:
            logAssert(
                np.isclose(r._h, 47.84, 1e-2),
                "h {} is wrong for couple {}".format(r._h, couple),
            )
            logAssert(
                np.isclose(r._r, 8.43, 1e-2),
                "R {} is wrong for couple {}".format(r._r, couple),
            )

        # Closed tops, closed bottoms
        elif couple[0] in ffr.CLOSED_TOPS and couple[1] in ffr.ROUND_BOTS:
            logAssert(
                np.isclose(r._h, 36.61, 1e-2),
                "h {} is wrong for couple {}".format(r._h, couple),
            )
            logAssert(
                np.isclose(r._r, 8.43, 1e-2),
                "R {} is wrong for couple {}".format(r._r, couple),
            )


def test_valid_stl(all_reactors: list):

    """
    Generate all possible shapes of reactor, and render the STL files.
    Then test the validity of each STL file:
    - Check if watertight (True if there is non non-manifold surfaces)
    - Check if valid volume
    - Check if valid winding
    """

    l.debug("Testing STLs validity")

    for couple in all_reactors:

        # Pass threaded tops, radius is constrained by the filter
        if couple[0] in rea.THREADED_TOPS:
            continue

        l.debug(
            "Building and rendering couple Top: {}; Bottom: {}".format(
                couple[0], couple[1]
            )
        )

        # Build the reactor
        r = ffr(6, 4, couple[0], couple[1], 20, 3)

        # Add inputs to check for non-manifolds
        r.addInputPercentageBottom("i1", 0)
        r.addInputPercentageBottom("i2", 100)
        r.addInputPercentageTop("i3", 0)
        r.addInputPercentageTop("i4", 100)

        # Forge the name of the file
        name = "{} {}".format(couple[0], couple[1])
        name = name.replace(" ", "_")
        name = name.replace("/", "_")
        r.renderToFile(name + ".scad")

        # Render the reactor into a stl file
        cmd = sp.Popen(
            ["openscad", "-o", name + ".stl", name + ".scad"],
            stdout=sp.PIPE,
            stderr=sp.PIPE,
        )

        # Wait until rendering is finished
        out, err = cmd.communicate()

        # Check if rendering succeeded, crash the test otherwise
        if cmd.returncode != 0:
            l.error(
                "Failed to render STL for couple Top {}; Bottom: {}".format(
                    couple[0], couple[1]
                )
            )
            pytest.fail(
                "Failed to render STL for couple Top: {}; Bottom: {}".format(
                    couple[0], couple[1]
                )
            )

        # Generate the mesh from STL file
        mesh = trimesh.load_mesh(name + ".stl")

        logAssert(
            mesh.is_watertight,
            "Couple Top: {}; Bottom: {} has non-manifold surfaces".format(
                couple[0], couple[1]
            ),
        )

        logAssert(
            mesh.is_volume,
            "Couple Top: {}; Bottom: {} has a negative volume".format(
                couple[0], couple[1]
            ),
        )

        logAssert(
            mesh.is_winding_consistent,
            "Couple Top: {}; Bottom: {} has unconsistent winding".format(
                couple[0], couple[1]
            ),
        )

        # Remove files to not polute dir
        os.remove(name + ".scad")
        os.remove(name + ".stl")


if __name__ == "__main__":
    pass



import os
import pytest
import numpy as np
import itertools
import subprocess as sp
import trimesh
import solid.utils as su

from ccad.log import MyLog
from ccad.reactor import ReactorCAD as rea
from ccad.tube_connector import TubeConnectorCAD
from ccad.exceptions import *

l = MyLog("output_tests_siphon.log", mode="w")

l.debug("---------------------- START NEW RUN OF TESTS ----------------------")
l.debug("\n")


def logAssert(test, msg):

    """
    Function to log the result of an assert
    http://stackoverflow.com/questions/24892396/py-test-logging-messages-and-test-results-assertions-into-a-single-file
    """

    if not test:
        l.error(msg)
        assert test, msg


def test_output_constraint():

    """Crash if output too high"""

    l.debug("Testing constraints on outputs")

    r_a = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)
    r_b = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)

    r_a.addOutputPercentage("test", height_per=50)
    r_b.addInputPercentage("test", height_per=50)

    t = TubeConnectorCAD(r_a, r_a.outputs["test"], r_b, r_b.inputs["test"])


def test_vertical_translation():
    l.debug("Testing vertical translations")
    r_a = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)
    r_b = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)

    r_b.addInputPercentage("test", height_per=100)

    t = TubeConnectorCAD(r_a, r_a.outputs["default"], r_b, r_b.inputs["test"], conflicts="lift_in_obj")

    logAssert(
        np.isclose(np.array(r_a.coo["z"]), 40.55, 1e-2),
        "Z does not match if lift in obj!"
    )  

    r_a = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)
    r_b = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)

    r_b.addInputPercentage("test", height_per=0)
    r_a.addOutputPercentage("test", height_per=100)

    t = TubeConnectorCAD(r_a, r_a.outputs["test"], r_b, r_b.inputs["test"], conflicts="lift_out_obj")

    logAssert(
        np.isclose(np.array(r_b.coo["z"]), 34.55, 1e-2),
        "Z does not match if lift out obj!"
    )  


def test_properties():
    """Test Length, width, and Height of TC
    """
    l.debug("Testing properties")

    r_a = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)
    r_b = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)

    r_a.addOutputPercentage("test", height_per=50)
    r_b.addInputPercentage("test", height_per=50)

    t = TubeConnectorCAD(r_a, r_a.outputs["test"], r_b, r_b.inputs["test"])

    logAssert(
        np.isclose(t.length, 1.33, 1e-2),
        "Properties: Wrong length: {}".format(t.length),
    )

    logAssert(t.width == 9, "Properties: Wrong width: {}".format(t.width))

    logAssert(
        np.isclose(t.height, 9.00, 1e-2),
        "Properties: Wrong height: {}".format(t.height),
    )

    t.length = 10
    logAssert(t.length == 10, "Properties: Wrong length: {}".format(t.length))

    x = np.array(r_b.coo["x"])
    logAssert(
        np.isclose(x, -39.72, 1e-2),
        "Worng x after modification of reactor 2: {}".format(x),
    )


def test_translations():

    """
    Check that the siphon will properly translate the objects it connects
    """

    l.debug("Testing Translations")
    r_a = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)
    r_b = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)

    r_b.addInputPercentage("test", height_per=50)
    t = TubeConnectorCAD(r_a, r_a.outputs["default"], r_b, r_b.inputs["test"])
    x = np.array(r_b.coo["x"])

    logAssert(
        np.isclose(x, 31.06, 1e-2), "Translation: Wrong x for 2nd reactor: {}".format(x)
    )

    r_a = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O, 30)
    r_b = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)

    r_b.addInputPercentage("test", height_per=50)
    t = TubeConnectorCAD(r_a, r_a.outputs["default"], r_b, r_b.inputs["test"])

    x = np.array(r_b.coo["x"])
    y = np.array(r_b.coo["y"])

    logAssert(
        np.isclose(x, 26.89, 1e-2),
        "Translation: Wrong x {} for 2nd reactor, output 1st at 30deg".format(x),
    )

    logAssert(
        np.isclose(y, 15.53, 1e-2),
        "Translation: Wrong y {} for 2nd reactor, output 1st at 30deg".format(y),
    )

    # 3 reactors, first one has its output at 30°, 2nd one has its input at
    # -30° (30 ° from its output). Results in collision between 1st and 3rd
    # reactor
    r_a = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O, 30)
    r_b = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O, 0)
    r_c = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O, 0)

    r_b.addInputPercentage("test", height_per=50, angle=-30)
    r_c.addInputPercentage("test", height_per=50)

    t1 = TubeConnectorCAD(r_a, r_a.outputs["default"], r_b, r_b.inputs["test"])
    t2 = TubeConnectorCAD(r_b, r_b.outputs["default"], r_c, r_c.inputs["test"])

    x = np.array(r_c.coo["x"])
    y = np.array(r_c.coo["y"])
    angle = np.array(r_c.coo["angle"])

    logAssert(
        np.isclose(x, 11.36, 1e-2), "Translation: Wrong x {} for 3rd reactor".format(x)
    )

    logAssert(
        np.isclose(y, -11.36, 1e-2), "Translation: Wrong y {} for 3rd reactor".format(y)
    )

    logAssert(np.isclose(angle, 240, 1e-2), "Translation: Wrong angle {}".format(angle))


def test_valid_stl():

    """
    Render the STL file for 2 reactors, one siphon
    Then test the validity of the STL file:
    - Check if watertight (True if there is non non-manifold surfaces)
    - Check valid volume
    """

    l.debug("Testing STLs validity")

    r_a = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)
    r_b = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)

    r_b.addInputPercentage("test", height_per=50)
    r_a.addOutputPercentage("test", height_per=50)

    c = TubeConnectorCAD(r_a, r_a.outputs["test"], r_b, r_b.inputs["test"])
    c.length = 5

    su.scad_render_to_file(r_a.cad + r_b.cad + c.cad, "test.scad")

    # Render the reactor into a stl file
    cmd = sp.Popen(
        ["openscad", "-o", "test.stl", "test.scad"], stdout=sp.PIPE, stderr=sp.PIPE
    )

    # Wait until rendering is finished
    out, err = cmd.communicate()

    # Check if rendering succeeded, crash the test otherwise
    if cmd.returncode != 0:
        l.error("Failed to render STL for 2 reactors, 1 siphon")
        pytest.fail("Failed to render STL for 2 reactors, 1 siphon")

    # Generate the mesh from STL file
    mesh = trimesh.load_mesh("test.stl")

    logAssert(mesh.is_watertight is True, "2 reactors, 1 siphon, not watertight")

    logAssert(mesh.volume > 0, "2 reactors, 1 siphon, negative volume")

    # Remove files to not polute dir
    os.remove("test.scad")
    os.remove("test.stl")



#!/usr/bin/python
# coding: utf-8


import sys
import os
from PyQt5 import QtGui, QtCore, QtWidgets

from ccad.in_out import InputOutput
import ccad.constants as cst


class AddTopIODialogFFR(QtWidgets.QDialog):

    """
    Dialog box opened from module_reactor when the user adds/modify a top inlet
    """

    def __init__(self, parent=None, io_dict=None):

        super(AddTopIODialogFFR, self).__init__(parent)

        self.setModal(True)

        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

        self.parent = parent

        self.io_dict = io_dict

        if parent is None:
            sys.path.append(
                os.path.dirname(
                    os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
                )
            )
            from log import MyLog

            self.l = MyLog("activity.log")

            # Dummy file for saving if testing
            # self.options = QtCore.QSettings("debug/options.ini",
            # QtCore.QSettings.IniFormat)

            self.test = True

            self.TOP_LUER = "Luer top inlet"
            self.TOP_CUSTOM = "Custom top inlet"
        else:
            self.l = self.parent.l
            # self.options = self.parent.options
            self.test = False

            self.TOP_LUER = self.parent.TOP_LUER
            self.TOP_CUSTOM = self.parent.TOP_CUSTOM

        self.l.debug("AddTopIODialog opened")

        self.initUI()
        self.defineSlots()

        # Dict to store module's features
        self.params = dict()
        self.restoreParams()

    def defineSlots(self):

        """Establish the slots"""

        # To close the window and build the module
        self.ok_button.clicked.connect(self.buildIO)

        # I/O type changed
        self.combo_type_io.currentIndexChanged.connect(self.typeIOChanged)

    def typeIOChanged(self):

        """Slot called when type of top I/O (Luer, custom, etc) is changed"""

        # Disable diameter field if top I/O is Luer
        if self.combo_type_io.currentText() == self.TOP_LUER:
            self.spin_diameter.setEnabled(False)
            self.spin_length.setEnabled(False)
            self.spin_diameter.setValue(cst.D_EXT_INLET)
            self.spin_length.setValue(cst.L_EXT_INLET)
        elif self.combo_type_io.currentText() == self.TOP_CUSTOM:
            self.spin_diameter.setEnabled(True)
            self.spin_length.setEnabled(True)

    def restoreParams(self):

        """
        If a dict is provided to the class, we are updating the object.
        Restore the parameters previously entered to the display
        """

        # If no io_dict, we're creating an input, nothing to restore
        if self.io_dict is None:
            return

        # Get the values of the object from self.io_dict
        # These 2 parameters are injected by the parent in the io_dict
        self.params["name"] = self.io_dict["name"]
        self.params["type_io"] = self.io_dict["type_io"]

        # Get the values of the object from self.io_dict
        self.params["diameter"] = self.io_dict["diameter"]
        self.params["walls"] = self.io_dict["walls"]
        self.params["length"] = self.io_dict["length"]

        # Set values for graphical elements
        # Disable name editing
        self.line_name.setText(self.params["name"])
        self.line_name.setEnabled(False)

        # Restore type of I/O to combo box
        if self.params["type_io"] == "luer":
            self.combo_type_io.setCurrentIndex(0)
            self.spin_diameter.setEnabled(False)
        else:
            self.combo_type_io.setCurrentIndex(1)

        # Disable combo box for type of I/O, user cant' change it
        self.combo_type_io.setEnabled(False)

        self.spin_diameter.setValue(self.params["diameter"])
        self.spin_walls.setValue(self.params["walls"])
        self.spin_length.setValue(self.params["length"])

        # TODO: handle check box for auto-placement

        # 'external' is a bool, convert it to text for display in combo box
        # if self.params['external']:
        # self.combo_external.setCurrentText('External')
        # else:
        # self.combo_external.setCurrentText('Internal')

        # self.spin_angle.setValue(self.params['angle'])

    def readForm(self):

        """
        Read the form, and get the parameters of the I/O. Put them in a dict
        """

        self.params["name"] = self.line_name.text()
        self.params["type_io"] = self.combo_type_io.currentText()
        self.params["diameter"] = self.spin_diameter.value()
        self.params["walls"] = self.spin_walls.value()
        self.params["length"] = self.spin_length.value()

        # Transform combo box text to valid value for TopInlet object
        if self.params["type_io"] == self.TOP_LUER:
            self.params["type_io"] = "luer"
        else:
            self.params["type_io"] = "custom"

        # Get the state of the auto_placement check box
        auto_placement = self.check_placement.checkState()
        if auto_placement == 2:
            auto_placement = True
        else:
            auto_placement = False

        self.params["auto_placement"] = auto_placement

    def _checkIOName(self, name: str) -> bool:

        """
        Check io's name is valid (not already used by otherI/O), otherwise
        display error dialog box
        """

        # If updating I/O, name is valid
        if self.io_dict is not None:
            return True

        mes = """
        An I/O with this name already exists for this reactor.\n
         Please use another name
        """

        mes2 = """
        You can't use an empty name for an I/O.\n
         Please give a name to this I/O
        """

        mes = mes.replace("    ", "")

        # Build a list of all I/O names
        names = (
            list(self.parent.dict_inputs)
            + list(self.parent.dict_outputs)
            + list(self.parent.dict_top_inlets)
        )

        if name in names:

            self.l.debug("Invalid name for I/O, display error dialog")
            QtWidgets.QMessageBox.critical(
                self, "Naming error", mes, QtWidgets.QMessageBox.Ok
            )

            return False

        if not name:

            self.l.debug("Invalid (empty) name for I/O, display error dialog")
            QtWidgets.QMessageBox.critical(
                self, "Naming error", mes2, QtWidgets.QMessageBox.Ok
            )

            return False

        return True

    def buildIO(self):

        """
        Called when user clicks the OK button. Will read the form  and call the
        parent to add the I/O
        """

        self.readForm()

        # If name is not valid, don't create I/O
        if not self._checkIOName(self.params["name"]):
            return

        if not self.test:
            self.parent.addTopInlet(self.params)

        # Close the settings window and free the memory
        self.close()

    def initUI(self):

        """Handles the display"""

        if self.io_dict is None:
            self.setWindowTitle("Add a top I/O")
        else:
            self.setWindowTitle("Modify a top I/O")

        # ----------------- BASIC SETTINGS ------------------------------------

        widget_form_basic = QtWidgets.QWidget(self)
        fbox = QtWidgets.QFormLayout()
        widget_form_basic.setLayout(fbox)

        # Line edit for input's name
        self.line_name = QtWidgets.QLineEdit(self)

        # Combo box to choose input or output
        self.combo_type_io = QtWidgets.QComboBox(self)
        self.combo_type_io.addItems([self.TOP_LUER, self.TOP_CUSTOM])

        # Diameter
        self.spin_diameter = QtWidgets.QDoubleSpinBox(self)
        self.spin_diameter.setMinimum(0)

        # Disable diameter spin box cause combo_type_io is Luer when
        # creating window. Set diameter value for Luer inlet
        self.spin_diameter.setValue(cst.D_EXT_INLET)
        self.spin_diameter.setEnabled(False)

        # Length of inlet
        self.spin_length = QtWidgets.QDoubleSpinBox(self)
        self.spin_length.setMinimum(0)

        # Freeze length field if I/O is Luer
        self.spin_length.setValue(cst.L_EXT_INLET)
        self.spin_length.setEnabled(False)

        # Walls
        self.spin_walls = QtWidgets.QDoubleSpinBox(self)
        self.spin_walls.setMinimum(0)

        # Set walls to default value since Luer inlet is selected when
        # creating window, but don't disable field so user can adapt value
        # easily
        self.spin_walls.setValue(cst.WALLS)

        self.check_placement = QtWidgets.QCheckBox(self)
        self.check_placement.setCheckState(2)

        # # TODO: TEMPORARY value
        # self.spin_diameter.setValue(3)

        # Add widgets to form layout
        fbox.addRow("Name: ", self.line_name)
        fbox.addRow("Type: ", self.combo_type_io)
        fbox.addRow("Diameter: ", self.spin_diameter)
        fbox.addRow("Length: ", self.spin_length)
        fbox.addRow("Walls: ", self.spin_walls)
        fbox.addRow("Auto placement: ", self.check_placement)

        # ----------------- ASSEMBLING ----------------------------------------

        self.vbox_global = QtWidgets.QVBoxLayout(self)

        self.ok_button = QtWidgets.QPushButton("OK", self)

        self.vbox_global.addWidget(widget_form_basic)
        self.vbox_global.addWidget(self.ok_button)

        self.setLayout(self.vbox_global)
        self.show()


if __name__ == "__main__":
    print(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
    app = QtWidgets.QApplication(sys.argv)
    obj = AddTopIODialogFFR()
    sys.exit(app.exec_())




import os
import pytest
import numpy as np
import subprocess as sp
import trimesh
import solid.utils as su

from ccad.log import MyLog
from ccad.reactor import ReactorCAD as rea
from ccad.double_siphon import DoubleSiphonCAD
from ccad.exceptions import *

l = MyLog("output_tests_siphon.log", mode="w")

l.debug("---------------------- START NEW RUN OF TESTS ----------------------")
l.debug("\n")


def logAssert(test, msg):

    """
    Function to log the result of an assert
    http://stackoverflow.com/questions/24892396/py-test-logging-messages-and-test-results-assertions-into-a-single-file
    """

    if not test:
        l.error(msg)
        assert test, msg


def test_output_height_constraint():

    """
    Crash if output too high
    """
    l.debug("Testing height constraint")

    r_a = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)
    r_b = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)

    r_a.addOutputPercentage("test", height_per=10)
    r_b.addInputPercentage("test")

    with pytest.warns(UserWarning):
        d = DoubleSiphonCAD(r_a, r_a.outputs["test"], r_b, r_b.inputs["test"])

    logAssert(
        np.isclose(d._height, 49.55, 1e-2), "Wrong height for DS after adapting height"
    )


def test_properties():

    """Test Length, width, and height of DS
    """
    l.debug("Testing properties")

    r_a = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)
    r_b = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)

    r_a.addOutputPercentage("test", height_per=20)
    r_b.addInputPercentage("test")

    d = DoubleSiphonCAD(r_a, r_a.outputs["test"], r_b, r_b.inputs["test"])

    logAssert(
        np.isclose(d.length, 16.33, 1e-2),
        "Properties: Wrong length: {}".format(d.length),
    )

    logAssert(d.width == 9, "Properties: Wrong width: {}".format(d.width))

    logAssert(
        np.isclose(d.height, 49.55, 1e-2),
        "Properties: Wrong height: {}".format(d.height),
    )

    d.length = 30
    logAssert(d.length == 30, "Wrong length: {}".format(d.length))

    x = np.array(r_b.coo["x"])
    logAssert(
        np.isclose(x, -59.72, 1e-2),
        "Wrong x after modification for reactor 2: {}".format(x),
    )


def test_translations():

    """
    Check that the siphon will properly translate the objects it connects
    """
    l.debug("Testing Translations")
    r_a = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)
    r_b = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)

    r_b.addInputPercentage("test")

    c = DoubleSiphonCAD(r_a, r_a.outputs["default"], r_b, r_b.inputs["test"])
    x = np.array(r_b.coo["x"])

    logAssert(
        np.isclose(x, 46.06, 1e-2), "Translation: Wrong x for 2nd reactor: {}".format(x)
    )

    # 2 reactors, first one has its output at 30°
    r_a = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O, 30)
    r_b = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)

    r_b.addInputPercentage("test")

    c = DoubleSiphonCAD(r_a, r_a.outputs["default"], r_b, r_b.inputs["test"])

    x = np.array(r_b.coo["x"])
    y = np.array(r_b.coo["y"])

    logAssert(
        np.isclose(x, 39.88, 1e-2),
        "Translation: Wrong x {} for 2nd reactor, output 1st at 30deg".format(x),
    )

    logAssert(
        np.isclose(y, 23.03, 1e-2),
        "Translation: Wrong y {} for 2nd reactor, output 1st at 30deg".format(y),
    )

    # 3 reactors, first one has its output at 30°, 2nd one has its input at
    # -30° (30 ° from its output). Results in collision between 1st and 3rd
    # reactor
    r_a = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O, 30)
    r_b = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O, 0)
    r_c = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O, 0)

    r_b.addInputPercentage("test", angle=-30)
    r_c.addInputPercentage("test")

    c = DoubleSiphonCAD(r_a, r_a.outputs["default"], r_b, r_b.inputs["test"])
    c_2 = DoubleSiphonCAD(r_b, r_b.outputs["default"], r_c, r_c.inputs["test"])

    x = np.array(r_c.coo["x"])
    y = np.array(r_c.coo["y"])
    angle = np.array(r_c.coo["angle"])

    logAssert(
        np.isclose(x, 16.85, 1e-2), "Translation: Wrong x {} for 3rd reactor".format(x)
    )

    logAssert(
        np.isclose(y, -16.85, 1e-2), "Translation: Wrong y {} for 3rd reactor".format(y)
    )

    logAssert(np.isclose(angle, 240, 1e-2), "Translation: Wrong angle {}".format(angle))


def test_valid_stl():

    """
    Render the STL file for 2 reactors, one siphon
    Then test the validity of the STL file:
    - Check if watertight (True if there is non non-manifold surfaces)
    - Check valid volume
    """

    l.debug("Testing STLs validity")

    r_a = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)
    r_b = rea(20, rea.T_OPEN, rea.B_FLAT_IN_O)

    r_b.addInputPercentage("test")

    c = DoubleSiphonCAD(r_a, r_a.outputs["default"], r_b, r_b.inputs["test"])

    su.scad_render_to_file(r_a.cad + r_b.cad + c.cad, "test.scad")

    # Render the reactor into a stl file
    cmd = sp.Popen(
        ["openscad", "-o", "test.stl", "test.scad"], stdout=sp.PIPE, stderr=sp.PIPE
    )

    # Wait until rendering is finished
    out, err = cmd.communicate()

    # Check if rendering succeeded, crash the test otherwise
    if cmd.returncode != 0:
        l.error("Failed to render STL for 2 reactors, 1 siphon")
        pytest.fail("Failed to render STL for 2 reactors, 1 siphon")

    # Generate the mesh from STL file
    mesh = trimesh.load_mesh("test.stl")

    logAssert(mesh.is_watertight is True, "2 reactors, 1 siphon, not watertight")

    logAssert(mesh.volume > 0, "2 reactors, 1 siphon, negative volume")

    # Remove files to not polute dir
    os.remove("test.scad")
    os.remove("test.stl")
